code = '''
import json

import os, importlib

import numpy as np
import arcpy


def get_available_device(max_memory=0.8):
    """
    select available device based on the memory utilization status of the device
    :param max_memory: the maximum memory utilization ratio that is considered available
    :return: GPU id that is available, -1 means no GPU is available/uses CPU, if GPUtil package is not installed, will
    return 0 
    """
    try:
        import GPUtil
    except ModuleNotFoundError:
        return 0

    GPUs = GPUtil.getGPUs()
    freeMemory = 0
    available = 0
    for GPU in GPUs:
        if GPU.memoryUtil > max_memory:
            continue
        if GPU.memoryFree >= freeMemory:
            freeMemory = GPU.memoryFree
            available = GPU.id

    return available


features = {
    "displayFieldName": "",
    "fieldAliases": {"FID": "FID", "Class": "Class", "Confidence": "Confidence"},
    "geometryType": "esriGeometryPolygon",
    "fields": [
        {"name": "FID", "type": "esriFieldTypeOID", "alias": "FID"},
        {"name": "Class", "type": "esriFieldTypeString", "alias": "Class"},
        {"name": "Confidence", "type": "esriFieldTypeDouble", "alias": "Confidence"},
    ],
    "features": [],
}

fields = {
    "fields": [
        {"name": "OID", "type": "esriFieldTypeOID", "alias": "OID"},
        {"name": "Class", "type": "esriFieldTypeString", "alias": "Class"},
        {"name": "Confidence", "type": "esriFieldTypeDouble", "alias": "Confidence"},
        {"name": "Shape", "type": "esriFieldTypeGeometry", "alias": "Shape"},
    ]
}


class GeometryType:
    Point = 1
    Multipoint = 2
    Polyline = 3
    Polygon = 4


class ArcGISObjectDetector:
    def __init__(self):
        self.name = "Object Detector"
        self.description = "This python raster function applies deep learning model to detect objects in imagery"

    def initialize(self, **kwargs):
        if "model" not in kwargs:
            return

        model = kwargs["model"]
        model_as_file = True
        try:
            with open(model, "r") as f:
                self.json_info = json.load(f)
        except FileNotFoundError:
            try:
                self.json_info = json.loads(model)
                model_as_file = False
            except json.decoder.JSONDecodeError:
                raise Exception("Invalid model argument")

        framework = self.json_info["Framework"]
        if "ModelConfiguration" in self.json_info:
            if isinstance(self.json_info["ModelConfiguration"], str):
                ChildModelDetector = getattr(
                    importlib.import_module(
                        "{}.{}".format(framework, self.json_info["ModelConfiguration"])
                    ),
                    "ChildObjectDetector",
                )
            else:
                ChildModelDetector = getattr(
                    importlib.import_module(
                        "{}.{}".format(
                            framework, self.json_info["ModelConfiguration"]["Name"]
                        )
                    ),
                    "ChildObjectDetector",
                )
        else:
            raise Exception("Invalid model configuration")

        os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"
        device = None
        if "device" in kwargs:
            device = kwargs["device"]
            if device == -2:
                device = get_available_device()

        if device is not None:
            if device >= 0:
                try:
                    import torch
                except Exception:
                    raise Exception(
                        "PyTorch is not installed. Install it using conda install -c esri deep-learning-essentials"
                    )
                torch.cuda.set_device(device)
                arcpy.env.processorType = "GPU"
                arcpy.env.gpuId = str(device)
            else:
                arcpy.env.processorType = "CPU"

        self.child_object_detector = ChildModelDetector()
        self.child_object_detector.initialize(model, model_as_file)

    def getParameterInfo(self):
        required_parameters = [
            {
                "name": "raster",
                "dataType": "raster",
                "required": True,
                "displayName": "Raster",
                "description": "Input Raster",
            },
            {
                "name": "model",
                "dataType": "string",
                "required": True,
                "displayName": "Input Model Definition (EMD) File",
                "description": "Input model definition (EMD) JSON file",
            },
            {
                "name": "device",
                "dataType": "numeric",
                "required": False,
                "displayName": "Device ID",
                "description": "Device ID",
            },
        ]
        parameter_info = self.child_object_detector.getParameterInfo(
            required_parameters
        )
        parameter_info.extend(
            [
                {
                    "name": "test_time_augmentation",
                    "dataType": "string",
                    "required": False,
                    "value": "False"
                    if "test_time_augmentation" not in self.json_info
                    else str(self.json_info["test_time_augmentation"]),
                    "displayName": "Perform test time augmentation while predicting",
                    "description": "If True, will merge predictions from flipped and rotated images.",
                }
            ]
        )
        return parameter_info

    def getConfiguration(self, **scalars):
        configuration = self.child_object_detector.getConfiguration(**scalars)
        if "DataRange" in self.json_info:
            configuration["dataRange"] = tuple(self.json_info["DataRange"])
        configuration["inheritProperties"] = 2 | 4 | 8
        configuration["inputMask"] = True
        self.use_tta = scalars.get("test_time_augmentation", "false").lower() in [
            "true",
            "1",
            "t",
            "y",
            "yes",
        ]
        self.nms_overlap = float(scalars.get("nms_overlap", 0.1))
        return configuration

    def getFields(self):
        return json.dumps(fields)

    def getGeometryType(self):
        return GeometryType.Polygon

    def vectorize(self, **pixelBlocks):
        # set pixel values in invalid areas to 0
        raster_mask = pixelBlocks["raster_mask"]
        raster_pixels = pixelBlocks["raster_pixels"]
        raster_pixels[np.where(raster_mask == 0)] = 0
        pixelBlocks["raster_pixels"] = raster_pixels

        polygon_list, scores, classes = self.tta_detect_objects(**pixelBlocks)

        features["features"] = []
        for i in range(len(polygon_list)):
            rings = [[]]
            for j in range(polygon_list[i].shape[0]):
                rings[0].append([polygon_list[i][j][1], polygon_list[i][j][0]])

            features["features"].append(
                {
                    "attributes": {
                        "OID": i + 1,
                        "Class": self.json_info["Classes"][classes[i] - 1]["Name"],
                        "Confidence": scores[i],
                    },
                    "geometry": {"rings": rings},
                }
            )

        return {"output_vectors": json.dumps(features)}

    def tta_detect_objects(self, **pixelBlocks):
        import torch
        from fastai.vision.transform import dihedral_affine, rotate
        from fastai.vision import Image

        input_image = pixelBlocks["raster_pixels"].astype(np.float32)

        tile_size = input_image.shape[1]
        pad = self.child_object_detector.padding

        allboxes = torch.empty(0,4)
        allclasses = []
        allscores = torch.empty(0)

        boxes_list, scores_list, labels_list = [], [], []
        transforms = [0]

        if self.use_tta:
            if self.json_info["ImageSpaceUsed"] == "MAP_SPACE":
                transforms = list(range(8))
            else:
                transforms = [
                    0,
                    2,
                ]  # no vertical flips for pixel space (oriented imagery)

        for k in transforms:
            out = dihedral_affine(Image(torch.tensor(input_image.copy() / 256.0)), k)
            pixelBlocksCopy = pixelBlocks.copy()
            pixelBlocksCopy["raster_pixels"] = (out.data * 256).numpy()
            polygons, scores, classes = self.detect_objects(**pixelBlocksCopy)

            bboxes = self.get_img_bbox(tile_size, polygons, scores, classes)
            if bboxes is not None:
                fixed_img_bboxes = dihedral_affine(bboxes, k)
                if k == 5 or k == 6:
                    fixed_img_bboxes = rotate(fixed_img_bboxes, 180)

                allboxes = torch.cat([allboxes, (fixed_img_bboxes.data[0]+1) / 2.0])
                allclasses = allclasses + fixed_img_bboxes.data[1].tolist()
                allscores = np.concatenate([allscores, torch.tensor(scores) * 0.01])

                boxes_list.append((fixed_img_bboxes.data[0] + 1) / 2.0)
                scores_list.append(torch.tensor(scores) * 0.01)  # normalize to [0,1]
                labels_list.append(fixed_img_bboxes.data[1].tolist())

        try:
            from ensemble_boxes import weighted_boxes_fusion

            iou_thr = self.nms_overlap
            skip_box_thr = 0.0001

            boxes, scores, labels = weighted_boxes_fusion(
                boxes_list,
                scores_list,
                labels_list,
                iou_thr=iou_thr,
                skip_box_thr=skip_box_thr,
            )
        except:
            import warnings

            warnings.warn("Unable to perform weighted boxes fusion... use NMS")
            boxes, scores, labels = np.array(allboxes), allscores, np.array(allclasses)

        bboxes = boxes * tile_size - pad
        polygons = self.convert_bounding_boxes_to_coord_list(bboxes)

        return polygons, np.array(scores * 100).astype(float), labels.astype(int)

    def get_img_bbox(self, tile_size, polygons, scores, classes):
        from fastai.vision import ImageBBox

        pad = self.child_object_detector.padding
        bboxes = []
        for i, polygon in enumerate(polygons):
            x1, y1 = np.around(polygon).astype(int)[0]
            x2, y2 = np.around(polygon).astype(int)[2]
            bboxes.append([x1 + pad, y1 + pad, x2 + pad, y2 + pad])
        n = len(bboxes)
        if n > 0:
            return ImageBBox.create(
                tile_size,
                tile_size,
                bboxes,
                labels=classes,
                classes=["Background"] + [x["Name"] for x in self.json_info["Classes"]],
            )
        else:
            return None

    def convert_bounding_boxes_to_coord_list(self, bounding_boxes):
        """
        convert bounding box numpy array to python list of point arrays
        :param bounding_boxes: numpy array of shape [n, 4]
        :return: python array of point numpy arrays, each point array is in shape [4,2]
        """
        num_bounding_boxes = bounding_boxes.shape[0]
        bounding_box_coord_list = []
        for i in range(num_bounding_boxes):
            coord_array = np.empty(shape=(4, 2), dtype=float)
            coord_array[0][0] = bounding_boxes[i][0]
            coord_array[0][1] = bounding_boxes[i][1]

            coord_array[1][0] = bounding_boxes[i][0]
            coord_array[1][1] = bounding_boxes[i][3]

            coord_array[2][0] = bounding_boxes[i][2]
            coord_array[2][1] = bounding_boxes[i][3]

            coord_array[3][0] = bounding_boxes[i][2]
            coord_array[3][1] = bounding_boxes[i][1]

            bounding_box_coord_list.append(coord_array)

        return bounding_box_coord_list

    def detect_objects(self, **pixelBlocks):
        polygon_list, scores, classes = self.child_object_detector.vectorize(
            **pixelBlocks
        )

        padding = self.child_object_detector.padding
        keep_polygon = []
        keep_scores = []
        keep_classes = []

        chip_sz = self.json_info["ImageHeight"]

        for idx, polygon in enumerate(polygon_list):
            centroid = polygon.mean(0)
            i, j = int(centroid[0]) // chip_sz, int(centroid[1]) // chip_sz
            x, y = int(centroid[0]) % chip_sz, int(centroid[1]) % chip_sz

            x1, y1 = polygon[0]
            x2, y2 = polygon[2]

            # fix polygon by removing padded regions
            polygon[:, 0] = polygon[:, 0] - (2 * i + 1) * padding
            polygon[:, 1] = polygon[:, 1] - (2 * j + 1) * padding

            X1, Y1, X2, Y2 = (
                i * chip_sz,
                j * chip_sz,
                (i + 1) * chip_sz,
                (j + 1) * chip_sz,
            )
            t = 2.0  # within 2 pixels of edge

            # if centroid not in center, reduce confidence
            # so box can be filtered out during NMS
            if (
                x < padding
                or x > chip_sz - padding
                or y < padding
                and y > chip_sz - padding
            ):

                scores[idx] = (self.child_object_detector.thres * 100) + scores[
                    idx
                ] * 0.01

            # if not excluded due to touching edge of tile
            if not (
                self.child_object_detector.filter_outer_padding_detections
                and any(
                    [
                        abs(X1 - x1) < t,
                        abs(X2 - x2) < t,
                        abs(Y1 - y1) < t,
                        abs(Y2 - y2) < t,
                    ]
                )
            ):  # touches edge
                keep_polygon.append(polygon)
                keep_scores.append(scores[idx])
                keep_classes.append(classes[idx])

        return keep_polygon, keep_scores, keep_classes


'''

feature_classifier_prf = """

import importlib
from importlib import reload, import_module
import json
import os
import sys
import arcpy
import numpy as np


def get_available_device(max_memory=0.8):
    '''
    select available device based on the memory utilization status of the device
    :param max_memory: the maximum memory utilization ratio that is considered available
    :return: GPU id that is available, -1 means no GPU is available/uses CPU, if GPUtil package is not installed, will
    return 0 
    '''
    try:
        import GPUtil
    except ModuleNotFoundError:
        return 0

    GPUs = GPUtil.getGPUs()
    freeMemory = 0
    available = 0
    for GPU in GPUs:
        if GPU.memoryUtil > max_memory:
            continue
        if GPU.memoryFree >= freeMemory:
            freeMemory = GPU.memoryFree
            available = GPU.id

    return available

features = {
    'displayFieldName': '',
    'fieldAliases': {
        'FID': 'FID',
        'Class': 'Class',
        'Confidence': 'Confidence'
    },
    'geometryType': 'esriGeometryPolygon',
    'fields': [
        {
            'name': 'FID',
            'type': 'esriFieldTypeOID',
            'alias': 'FID'
        },
        {
            'name': 'Class',
            'type': 'esriFieldTypeString',
            'alias': 'Class'
        },
        {
            'name': 'Confidence',
            'type': 'esriFieldTypeDouble',
            'alias': 'Confidence'
        }
    ],
    'features': []
}

fields = {
    'fields': [
        {
            'name': 'OID',
            'type': 'esriFieldTypeOID',
            'alias': 'OID'
        },
        {
            'name': 'Class',
            'type': 'esriFieldTypeString',
            'alias': 'Class'
        },
        {
            'name': 'Confidence',
            'type': 'esriFieldTypeDouble',
            'alias': 'Confidence'
        },
        {
            'name': 'Shape',
            'type': 'esriFieldTypeGeometry',
            'alias': 'Shape'
        }
    ]
}

class GeometryType:
    Point = 1
    Multipoint = 2
    Polyline = 3
    Polygon = 4


class ArcGISObjectClassifier:
    def __init__(self):
        self.name = 'Object classifier'
        self.description = 'This python raster function applies deep learning model to classify objects from overlaid imagery'

    def initialize(self, **kwargs):

        if 'model' not in kwargs:
            return

        # Read esri model definition (emd) file
        model = kwargs['model']
        model_as_file = True

        try:
            with open(model, 'r') as f:
                self.json_info = json.load(f)
        except FileNotFoundError:
            try:
                self.json_info = json.loads(model)
                model_as_file = False
            except json.decoder.JSONDecodeError:
                raise Exception("Invalid model argument")

        framework = self.json_info['Framework']
        if 'ModelConfiguration' in self.json_info:
            modelconfig = self.json_info['ModelConfiguration']
            if isinstance(modelconfig, str):
                if modelconfig not in sys.modules:
                    ChildModelDetector = getattr(import_module(
                        '{}.{}'.format(framework, modelconfig)), 'ChildObjectDetector')
                else:
                    ChildModelDetector = getattr(reload(
                        '{}.{}'.format(framework, modelconfig)), 'ChildObjectDetector')
            else:
                modelconfig = self.json_info['ModelConfiguration']['Name']
                ChildModelDetector = getattr(importlib.import_module(
                    '{}.{}'.format(framework, modelconfig)), 'ChildObjectDetector')
        else:
            raise Exception("Invalid model configuration")

        os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"

        device = None
        if 'device' in kwargs:
            device = kwargs['device']
            if device == -2:
                device = get_available_device()

        if device is not None:
            if device >= 0:
                try:
                    import torch
                except Exception:
                    raise Exception("PyTorch is not installed. Install it using conda install -c esri deep-learning-essentials")
                torch.cuda.set_device(device)
                arcpy.env.processorType = "GPU"
                arcpy.env.gpuId = str(device)
            else:
                arcpy.env.processorType = "CPU"

        self.child_object_detector = ChildModelDetector()
        self.child_object_detector.initialize(model, model_as_file)


    def getParameterInfo(self):

        # PRF needs values of these parameters from gp tool user,
        # either from gp tool UI or emd (a json) file.
        required_parameters = [
            {
                # To support mini batch, it is required that Classify Objects Using Deep Learning geoprocessing Tool
                # passes down a stack of raster tiles to PRF for model inference, the keyword required here is 'rasters'.
                'name': 'rasters',
                'dataType': 'rasters',
                'value': None,
                'required': True,
                'displayName': "Rasters",
                'description': 'The collection of overlapping rasters to objects to be classified'
            },
            {
                'name': 'model',
                'dataType': 'string',
                'required': True,
                'displayName': 'Input Model Definition (EMD) File',
                'description': 'Input model definition (EMD) JSON file'
            },
            {
                'name': 'device',
                'dataType': 'numeric',
                'required': False,
                'displayName': 'Device ID',
                'description': 'Device ID'
            }
        ]

        if 'BatchSize' not in self.json_info:
             required_parameters.append(
                 {
                     'name': 'batch_size',
                     'dataType': 'numeric',
                     'required': False,
                     'value': 4,
                     'displayName': 'Batch Size',
                     'description': 'Batch Size'
                 }
             )

        return self.child_object_detector.getParameterInfo(required_parameters)


    def getConfiguration(self, **scalars):


        # The information PRF returns to the GP tool,
        # the information is either from emd or defined in getConfiguration method.

        configuration = self.child_object_detector.getConfiguration(**scalars)

        if 'DataRange' in self.json_info:
            configuration['dataRange'] = tuple(self.json_info['DataRange'])

        configuration['inheritProperties'] = 2|4|8
        configuration['inputMask'] = True


        return configuration

    def getFields(self):

        fields = {
                'fields': [
                    {
                        'name': 'OID',
                        'type': 'esriFieldTypeOID',
                        'alias': 'OID'
                    },
                    {
                        'name': 'Class',
                        'type': 'esriFieldTypeString',
                        'alias': 'Class'
                    },
                    {
                        'name': 'Confidence',
                        'type': 'esriFieldTypeDouble',
                        'alias': 'Confidence'
                    },
                    {
                        'name': 'Shape',
                        'type': 'esriFieldTypeGeometry',
                        'alias': 'Shape'
                    }
                ]
            }
        fields['fields'].append(
            {
                'name': 'Label',
                'type': 'esriFieldTypeString',
                'alias': 'Label'
            }
        )

        if "MetaDataMode" in self.json_info and self.json_info["MetaDataMode"] == "MultiLabeled_Tiles":
            for item in fields['fields']:
                if item['name'] == 'Confidence':
                    item['type'] = 'esriFieldTypeString'

        return json.dumps(fields)

    def getGeometryType(self):
        return GeometryType.Polygon

    def vectorize(self, **pixelBlocks):

        # set pixel values in invalid areas to 0
        rasters_mask = pixelBlocks['rasters_mask']
        rasters_pixels = pixelBlocks['rasters_pixels']

        for i in range(0, len(rasters_pixels)):
            rasters_pixels[i][np.where(rasters_mask[i] == 0)] = 0

        pixelBlocks['rasters_pixels'] = rasters_pixels

        polygon_list, scores, labels = self.child_object_detector.vectorize(**pixelBlocks)

        features['features'] = []

        features['fieldAliases'].update({
            'Label':'Label'
        })

        features['fields'].append(
            {
                'name': 'Label',
                'type': 'esriFieldTypeString',
                'alias': 'Label'
            }
        )

        if "MetaDataMode" in self.json_info and self.json_info["MetaDataMode"] == "MultiLabeled_Tiles":
            for item in features['fields']:
                if item['name'] == 'Confidence':
                    item['type'] = 'esriFieldTypeString'

        for i in range(len(polygon_list)):

            rings = [[]]
            for j in range(len(polygon_list[i])):
                rings[0].append(
                    [
                        polygon_list[i][j][1],
                        polygon_list[i][j][0]
                    ]
                )

            features['features'].append({
                'attributes': {
                    'OID': i + 1,
                    'Confidence': str(scores[i]),
                    'Label': labels[i],
                    'Classname': labels[i]
                },
                'geometry': {
                    'rings': rings
                }
            })

        return {'output_vectors': json.dumps(features)}
"""
entity_recognizer_placeholder = """
print('not implemented')
"""

image_classifier_prf = """

import arcpy
import numpy as np
import json
import sys, os, importlib
import math


def get_available_device(max_memory=0.8):
    '''
    select available device based on the memory utilization status of the device
    :param max_memory: the maximum memory utilization ratio that is considered available
    :return: GPU id that is available, -1 means no GPU is available/uses CPU, if GPUtil package is not installed, will
    return 0
    '''
    try:
        import GPUtil
    except ModuleNotFoundError:
        return 0

    GPUs = GPUtil.getGPUs()
    freeMemory = 0
    available = 0
    for GPU in GPUs:
        if GPU.memoryUtil > max_memory:
            continue
        if GPU.memoryFree >= freeMemory:
            freeMemory = GPU.memoryFree
            available = GPU.id

    return available

def chunk_it(image, tile_size):
    s = image.shape
    num_rows = math.ceil(s[0]/tile_size)
    num_cols = math.ceil(s[1]/tile_size)
    r = np.array_split(image, num_rows)
    rows = []
    for x in r:
        x = np.array_split(x, num_cols, axis=1)
        rows.append(x)
    return rows, num_rows, num_cols

def crop_center(img, pad):
    if pad == 0:
        return img
    return img[pad:-pad, pad: -pad, :]

def crop_flatten(chunked, pad):
    imgs = []
    for r, row  in enumerate(chunked):
        for c, col in enumerate(row):
            col = crop_center(col, pad)
            imgs.append(col)
    return imgs

def patch_chips(imgs, n_rows, n_cols):
    h_stacks = []
    for i in range(n_rows):
        h_stacks.append(np.hstack(imgs[i*n_cols:n_cols*(i+1) ]))
    return np.vstack(h_stacks)

attribute_table = {
    'displayFieldName': '',
    'fieldAliases': {
        'OID': 'OID',
        'Value': 'Value',
        'Class': 'Class',
        'Red': 'Red',
        'Green': 'Green',
        'Blue': 'Blue'
    },
    'fields': [
        {
            'name': 'OID',
            'type': 'esriFieldTypeOID',
            'alias': 'OID'
        },
        {
            'name': 'Value',
            'type': 'esriFieldTypeInteger',
            'alias': 'Value'
        },
        {
            'name': 'Class',
            'type': 'esriFieldTypeString',
            'alias': 'Class'
        },
        {
            'name': 'Red',
            'type': 'esriFieldTypeInteger',
            'alias': 'Red'
        },
        {
            'name': 'Green',
            'type': 'esriFieldTypeInteger',
            'alias': 'Green'
        },
        {
            'name': 'Blue',
            'type': 'esriFieldTypeInteger',
            'alias': 'Blue'
        }
    ],
    'features': []
}

 

class ArcGISImageClassifier:
    def __init__(self):
        self.name = 'Image Classifier'
        self.description = 'Image classification python raster function to inference a pytorch image classifier'

    def initialize(self, **kwargs):
        if 'model' not in kwargs:
            return

        model = kwargs['model']
        model_as_file = True
        try:
            with open(model, 'r') as f:
                self.json_info = json.load(f)
        except FileNotFoundError:
            try:
                self.json_info = json.loads(model)
                model_as_file = False
            except json.decoder.JSONDecodeError:
                raise Exception("Invalid model argument")
        
        self.class_values = set([row['Value'] for row in self.json_info.get('Classes', [])])
        framework = self.json_info['Framework']
        if 'ModelConfiguration' in self.json_info:
            if isinstance(self.json_info['ModelConfiguration'], str):
                ChildImageClassifier = getattr(importlib.import_module(
                    '{}.{}'.format(framework, self.json_info['ModelConfiguration'])), 'ChildImageClassifier')
            else:
                ChildImageClassifier = getattr(importlib.import_module(
                    '{}.{}'.format(framework, self.json_info['ModelConfiguration']['Name'])), 'ChildImageClassifier')
        else:
            raise Exception("Invalid model configuration")

        os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"
        device = None
        if 'device' in kwargs:
            device = kwargs['device']
            if device == -2:
                device = get_available_device()

        if device is not None:
            if device >= 0:
                try:
                    import torch
                except Exception:
                    raise Exception("PyTorch is not installed. Install it using conda install -c esri deep-learning-essentials")
                torch.cuda.set_device(device)
                arcpy.env.processorType = "GPU"
                arcpy.env.gpuId = str(device)
            else:
                arcpy.env.processorType = "CPU"

        self.child_image_classifier = ChildImageClassifier()
        self.child_image_classifier.initialize(model, model_as_file)

    def getParameterInfo(self):
        required_parameters = [
            {
                'name': 'raster',
                'dataType': 'raster',
                'required': True,
                'displayName': 'Raster',
                'description': 'Input Raster'
            },
            {
                'name': 'model',
                'dataType': 'string',
                'required': True,
                'displayName': 'Input Model Definition (EMD) File',
                'description': 'Input model definition (EMD) JSON file'
            },
            {
                'name': 'device',
                'dataType': 'numeric',
                'required': False,
                'displayName': 'Device ID',
                'description': 'Device ID'
            }
        ]
        params = self.child_image_classifier.getParameterInfo(required_parameters)
        if 0 not in self.class_values and len(self.class_values)>1:
            for param in params:
                if param['name'] == 'predict_background':
                    param['value'] = 'False'
                    break
        return params

    def getConfiguration(self, **scalars):
        configuration = self.child_image_classifier.getConfiguration(**scalars)
        if 'DataRange' in self.json_info:
            configuration['dataRange'] = tuple(self.json_info['DataRange'])
        configuration['inheritProperties'] = 2|4|8
        configuration['inputMask'] = True
        return configuration

    def updateRasterInfo(self, **kwargs):
        kwargs['output_info']['bandCount'] = 1
        #todo: type is determined by the value range of classes in the json file
        prob_raster = getattr(self.child_image_classifier,'probability_raster',False)
        if prob_raster:
            kwargs['output_info']['pixelType'] = 'f4' # To ensure that output pixels are in prob range 0 to 1
        else:
            kwargs['output_info']['pixelType'] = 'i4'
        class_info = self.json_info['Classes']
        attribute_table['features'] = []
        for i, c in enumerate(class_info):
            attribute_table['features'].append(
                {
                    'attributes':{
                        'OID':i+1,
                        'Value':c['Value'],
                        'Class':c['Name'],
                        'Red':c['Color'][0],
                        'Green':c['Color'][1],
                        'Blue':c['Color'][2]
                    }
                }
            )
        kwargs['output_info']['rasterAttributeTable'] = json.dumps(attribute_table)

        return kwargs


    def updatePixels(self, tlc, shape, props, **pixelBlocks):
        # set pixel values in invalid areas to 0
           
        raster_mask = pixelBlocks['raster_mask']
        raster_pixels = pixelBlocks['raster_pixels']
        raster_pixels[np.where(raster_mask == 0)] = 0
        pixelBlocks['raster_pixels'] = raster_pixels

        if self.json_info['ModelName'] == 'MultiTaskRoadExtractor':
            xx = self.child_image_classifier.detectRoads(tlc, shape, props, **pixelBlocks).astype(props['pixelType'], copy=False)   
            pixelBlocks['output_pixels'] = xx
        elif hasattr(self.child_image_classifier, 'updatePixelsTTA'):
            xx = self.child_image_classifier.updatePixelsTTA(tlc, shape, props, **pixelBlocks).astype(props['pixelType'], copy=False)   
            pixelBlocks['output_pixels'] = xx
        else:
            xx = self.child_image_classifier.updatePixels(tlc, shape, props, **pixelBlocks).astype(props['pixelType'], copy=False)   
            tytx = getattr(self.child_image_classifier, 'tytx', self.json_info['ImageHeight'])
            chunks, num_rows, num_cols =  chunk_it(xx.transpose(1, 2, 0), tytx)# self.json_info['ImageHeight'])  # ImageHeight = ImageWidth
            xx = patch_chips(crop_flatten(chunks, self.child_image_classifier.padding), num_rows, num_cols)
            xx = xx.transpose(2, 0, 1)
            pixelBlocks['output_pixels'] = xx

        return pixelBlocks
"""

instance_detector_prf = """
import json
import sys, os, importlib

import numpy as np
import math
import arcpy

def get_centroid(polygon):
    polygon = np.array(polygon)
    return [polygon[:, 0].mean(), polygon[:, 1].mean()]        

def check_centroid_in_center(centroid, start_x, start_y, chip_sz, padding):
    return ((centroid[1] >= (start_y + padding)) and  \
                (centroid[1] <= (start_y + (chip_sz - padding))) and \
                (centroid[0] >= (start_x + padding)) and \
                (centroid[0] <= (start_x + (chip_sz - padding))))

def find_i_j(centroid, n_rows, n_cols, chip_sz, padding, filter_detections):
    for i in range(n_rows):
        for j in range(n_cols):
            start_x = i * chip_sz
            start_y = j * chip_sz

            if (centroid[1] > (start_y)) and (centroid[1] < (start_y + (chip_sz))) and (centroid[0] > (start_x)) and (centroid[0] < (start_x + (chip_sz))):
                in_center = check_centroid_in_center(centroid, start_x, start_y, chip_sz, padding)
                if filter_detections:
                    if in_center: 
                        return i, j, in_center
                else:
                    return i, j, in_center
    return None        

def get_available_device(max_memory=0.8):
    '''
    select available device based on the memory utilization status of the device
    :param max_memory: the maximum memory utilization ratio that is considered available
    :return: GPU id that is available, -1 means no GPU is available/uses CPU, if GPUtil package is not installed, will
    return 0 
    '''
    try:
        import GPUtil
    except ModuleNotFoundError:
        return 0

    GPUs = GPUtil.getGPUs()
    freeMemory = 0
    available = 0
    for GPU in GPUs:
        if GPU.memoryUtil > max_memory:
            continue
        if GPU.memoryFree >= freeMemory:
            freeMemory = GPU.memoryFree
            available = GPU.id

    return available

features = {
    'displayFieldName': '',
    'fieldAliases': {
        'FID': 'FID',
        'Class': 'Class',
        'Confidence': 'Confidence'
    },
    'geometryType': 'esriGeometryPolygon',
    'fields': [
        {
            'name': 'FID',
            'type': 'esriFieldTypeOID',
            'alias': 'FID'
        },
        {
            'name': 'Class',
            'type': 'esriFieldTypeString',
            'alias': 'Class'
        },
        {
            'name': 'Confidence',
            'type': 'esriFieldTypeDouble',
            'alias': 'Confidence'
        }
    ],
    'features': []
}

fields = {
    'fields': [
        {
            'name': 'OID',
            'type': 'esriFieldTypeOID',
            'alias': 'OID'
        },
        {
            'name': 'Class',
            'type': 'esriFieldTypeString',
            'alias': 'Class'
        },
        {
            'name': 'Confidence',
            'type': 'esriFieldTypeDouble',
            'alias': 'Confidence'
        },
        {
            'name': 'Shape',
            'type': 'esriFieldTypeGeometry',
            'alias': 'Shape'
        }
    ]
}

class GeometryType:
    Point = 1
    Multipoint = 2
    Polyline = 3
    Polygon = 4

class ArcGISInstanceDetector:
    def __init__(self):
        self.name = 'Instance Segmentation'
        self.description = 'Instance Segmentation python raster function to inference a arcgis.learn deep learning model.'

    def initialize(self, **kwargs):

        if 'model' not in kwargs:
            return

        model = kwargs['model']
        model_as_file = True
        try:
            with open(model, 'r') as f:
                self.json_info = json.load(f)
        except FileNotFoundError:
            try:
                self.json_info = json.loads(model)
                model_as_file = False
            except json.decoder.JSONDecodeError:
                raise Exception("Invalid model argument")

        framework = self.json_info['Framework']
        if 'ModelConfiguration' in self.json_info:
            if isinstance(self.json_info['ModelConfiguration'], str):
                ChildInstanceDetector = getattr(importlib.import_module(
                    '{}.{}'.format(framework, self.json_info['ModelConfiguration'])), 'ChildInstanceDetector')
            else:
                ChildInstanceDetector = getattr(importlib.import_module(
                    '{}.{}'.format(framework, self.json_info['ModelConfiguration']['Name'])), 'ChildInstanceDetector')
        else:
            raise Exception("Invalid model configuration")

        os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"
        device = None
        if 'device' in kwargs:
            device = kwargs['device']
            if device == -2:
                device = get_available_device()

        if device is not None:
            if device >= 0:
                try:
                    import torch
                except Exception:
                    raise Exception("PyTorch is not installed. Install it using conda install -c esri deep-learning-essentials")
                torch.cuda.set_device(device)
                arcpy.env.processorType = "GPU"
                arcpy.env.gpuId = str(device)
            else:
                arcpy.env.processorType = "CPU"

        self.child_instance_detector = ChildInstanceDetector()
        self.child_instance_detector.initialize(model, model_as_file)

        
    def getParameterInfo(self):       
        required_parameters = [
            {
                'name': 'raster',
                'dataType': 'raster',
                'required': True,
                'displayName': 'Raster',
                'description': 'Input Raster'
            },
            {
                'name': 'model',
                'dataType': 'string',
                'required': True,
                'displayName': 'Input Model Definition (EMD) File',
                'description': 'Input model definition (EMD) JSON file'
            },
            {
                'name': 'device',
                'dataType': 'numeric',
                'required': False,
                'displayName': 'Device ID',
                'description': 'Device ID'
            }
        ]     
        return self.child_instance_detector.getParameterInfo(required_parameters)


    def getConfiguration(self, **scalars):         
        configuration = self.child_instance_detector.getConfiguration(**scalars)
        if 'DataRange' in self.json_info:
            configuration['dataRange'] = tuple(self.json_info['DataRange'])
        configuration['inheritProperties'] = 2|4|8
        configuration['inputMask'] = True
        return configuration

    def getFields(self):
        return json.dumps(fields)

    def getGeometryType(self):          
        return GeometryType.Polygon        

    def vectorize(self, **pixelBlocks):
           
        raster_mask = pixelBlocks['raster_mask']
        raster_pixels = pixelBlocks['raster_pixels']
        raster_pixels[np.where(raster_mask == 0)] = 0
        pixelBlocks['raster_pixels'] = raster_pixels

        masks, pred_class, pred_score = self.child_instance_detector.vectorize(**pixelBlocks)

        n_rows = int(math.sqrt(self.child_instance_detector.batch_size))
        n_cols = int(math.sqrt(self.child_instance_detector.batch_size))
        padding = self.child_instance_detector.padding
        keep_masks = []
        keep_scores = []
        keep_classes = []       

        for idx, mask in enumerate(masks):
            if mask == []:
                continue
            centroid = get_centroid(mask[0])
            tytx = getattr(self.child_instance_detector, 'tytx', self.json_info['ImageHeight'])
            grid_location = find_i_j(centroid, n_rows, n_cols, tytx, padding, True)
            if grid_location is not None:
                i, j, in_center = grid_location
                for poly_id, polygon in enumerate(mask):
                    polygon = np.array(polygon)
                    polygon[:, 0] = polygon[:, 0] - (2*i + 1)*padding  # Inplace operation
                    polygon[:, 1] = polygon[:, 1] - (2*j + 1)*padding  # Inplace operation            
                    mask[poly_id] = polygon.tolist()
                if in_center:
                    keep_masks.append(mask)
                    keep_scores.append(pred_score[idx])
                    keep_classes.append(pred_class[idx])

        masks =  keep_masks
        pred_score = keep_scores
        pred_class = keep_classes        


        features['features'] = []

        for mask_idx, mask in enumerate(masks):

            features['features'].append({
                'attributes': {
                    'OID': mask_idx + 1,
                    'Class': self.json_info['Classes'][pred_class[mask_idx] - 1]['Name'],
                    'Confidence': pred_score[mask_idx]
                },
                'geometry': {
                    'rings': mask
                }
        }) 

        return {'output_vectors': json.dumps(features)}


"""
super_resolution = """

import arcpy
import numpy as np
import json
import sys, os, importlib
import math


def get_available_device(max_memory=0.8):
    '''
    select available device based on the memory utilization status of the device
    :param max_memory: the maximum memory utilization ratio that is considered available
    :return: GPU id that is available, -1 means no GPU is available/uses CPU, if GPUtil package is not installed, will
    return 0
    '''
    try:
        import GPUtil
    except ModuleNotFoundError:
        return 0

    GPUs = GPUtil.getGPUs()
    freeMemory = 0
    available = 0
    for GPU in GPUs:
        if GPU.memoryUtil > max_memory:
            continue
        if GPU.memoryFree >= freeMemory:
            freeMemory = GPU.memoryFree
            available = GPU.id

    return available

def chunk_it(image, tile_size):
    s = image.shape
    num_rows = math.ceil(s[0]/tile_size)
    num_cols = math.ceil(s[1]/tile_size)
    r = np.array_split(image, num_rows)
    rows = []
    for x in r:
        x = np.array_split(x, num_cols, axis=1)
        rows.append(x)
    return rows, num_rows, num_cols

def crop_center(img, pad):
    if pad == 0:
        return img
    return img[pad:-pad, pad: -pad, :]

def crop_flatten(chunked, pad):
    imgs = []
    for r, row  in enumerate(chunked):
        for c, col in enumerate(row):
            col = crop_center(col, pad)
            imgs.append(col)
    return imgs

def patch_chips(imgs, n_rows, n_cols):
    h_stacks = []
    for i in range(n_rows):
        h_stacks.append(np.hstack(imgs[i*n_cols:n_cols*(i+1) ]))
    return np.vstack(h_stacks)
 

class ArcGISSuperResolution:
    def __init__(self):
        self.name = 'Image Classifier'
        self.description = 'Image classification python raster function to inference a pytorch image classifier'

    def initialize(self, **kwargs):
        if 'model' not in kwargs:
            return

        model = kwargs['model']
        model_as_file = True
        try:
            with open(model, 'r') as f:
                self.json_info = json.load(f)
        except FileNotFoundError:
            try:
                self.json_info = json.loads(model)
                model_as_file = False
            except json.decoder.JSONDecodeError:
                raise Exception("Invalid model argument")

        framework = self.json_info['Framework']
        if 'ModelConfiguration' in self.json_info:
            if isinstance(self.json_info['ModelConfiguration'], str):
                ChildImageClassifier = getattr(importlib.import_module(
                    '{}.{}'.format(framework, self.json_info['ModelConfiguration'])), 'ChildImageClassifier')
            else:
                ChildImageClassifier = getattr(importlib.import_module(
                    '{}.{}'.format(framework, self.json_info['ModelConfiguration']['Name'])), 'ChildImageClassifier')
        else:
            raise Exception("Invalid model configuration")

        os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"
        device = None
        if 'device' in kwargs:
            device = kwargs['device']
            if device == -2:
                device = get_available_device()

        if device is not None:
            if device >= 0:
                try:
                    import torch
                except Exception:
                    raise Exception("PyTorch is not installed. Install it using conda install -c esri deep-learning-essentials")
                torch.cuda.set_device(device)
                arcpy.env.processorType = "GPU"
                arcpy.env.gpuId = str(device)
            else:
                arcpy.env.processorType = "CPU"

        self.child_image_classifier = ChildImageClassifier()
        self.child_image_classifier.initialize(model, model_as_file)

    def getParameterInfo(self):
        required_parameters = [
            {
                'name': 'raster',
                'dataType': 'raster',
                'required': True,
                'displayName': 'Raster',
                'description': 'Input Raster'
            },
            {
                'name': 'model',
                'dataType': 'string',
                'required': True,
                'displayName': 'Input Model Definition (EMD) File',
                'description': 'Input model definition (EMD) JSON file'
            },
            {
                'name': 'device',
                'dataType': 'numeric',
                'required': False,
                'displayName': 'Device ID',
                'description': 'Device ID'
            }
        ]
        return self.child_image_classifier.getParameterInfo(required_parameters)

    def getConfiguration(self, **scalars):
        configuration = self.child_image_classifier.getConfiguration(**scalars)
        if 'DataRange' in self.json_info:
            configuration['dataRange'] = tuple(self.json_info['DataRange'])
        #configuration['inheritProperties'] = 2|4|8
        #configuration['inputMask'] = True
        return configuration

    def updateRasterInfo(self, **kwargs):
        kwargs['output_info']['bandCount'] = self.json_info.get("n_channel", 3)
        kwargs['output_info']['pixelType'] = 'f4'
        return kwargs

    def updatePixels(self, tlc, shape, props, **pixelBlocks):
        # set pixel values in invalid areas to 0
           
        #raster_mask = pixelBlocks['raster_mask']
        raster_pixels = pixelBlocks['raster_pixels']
        #raster_pixels[np.where(raster_mask == 0)] = 0
        pixelBlocks['raster_pixels'] = raster_pixels

        if hasattr(self.child_image_classifier, "updatePixelsSmooth"):
            xx = self.child_image_classifier.updatePixelsSmooth( tlc, shape, props, **pixelBlocks).astype(props["pixelType"], copy=False)
            pixelBlocks["output_pixels"] = xx
        else:
            xx = self.child_image_classifier.updatePixels(tlc, shape, props, **pixelBlocks).astype(props['pixelType'], copy=False)
            chunks, num_rows, num_cols =  chunk_it(xx.transpose(1, 2, 0), self.json_info['ImageHeight'])  # ImageHeight = ImageWidth
            xx = patch_chips(crop_flatten(chunks, self.child_image_classifier.padding), num_rows, num_cols)
            xx = xx.transpose(2, 0, 1)
            pixelBlocks['output_pixels'] = xx

        return pixelBlocks

"""

image_translation_prf = """
import arcpy
import numpy as np
import json
import sys, os, importlib
import math

def get_available_device(max_memory=0.8):
    '''
    select available device based on the memory utilization status of the device
    :param max_memory: the maximum memory utilization ratio that is considered available
    :return: GPU id that is available, -1 means no GPU is available/uses CPU, if GPUtil package is not installed, will
    return 0
    '''
    try:
        import GPUtil
    except ModuleNotFoundError:
        return 0
    GPUs = GPUtil.getGPUs()
    freeMemory = 0
    available = 0
    for GPU in GPUs:
        if GPU.memoryUtil > max_memory:
            continue
        if GPU.memoryFree >= freeMemory:
            freeMemory = GPU.memoryFree
            available = GPU.id
    return available
def chunk_it(image, tile_size):
    s = image.shape
    num_rows = math.ceil(s[0]/tile_size)
    num_cols = math.ceil(s[1]/tile_size)
    r = np.array_split(image, num_rows)
    rows = []
    for x in r:
        x = np.array_split(x, num_cols, axis=1)
        rows.append(x)
    return rows, num_rows, num_cols
def crop_center(img, pad):
    if pad == 0:
        return img
    return img[pad:-pad, pad: -pad, :]
def crop_flatten(chunked, pad):
    imgs = []
    for r, row  in enumerate(chunked):
        for c, col in enumerate(row):
            col = crop_center(col, pad)
            imgs.append(col)
    return imgs
def patch_chips(imgs, n_rows, n_cols):
    h_stacks = []
    for i in range(n_rows):
        h_stacks.append(np.hstack(imgs[i*n_cols:n_cols*(i+1) ]))
    return np.vstack(h_stacks)
 
class ArcGISImageTranslation:
    def __init__(self):
        self.name = 'Image Classifier'
        self.description = 'Image classification python raster function to inference a pytorch image classifier'
    def initialize(self, **kwargs):
        if 'model' not in kwargs:
            return
        model = kwargs['model']
        model_as_file = True
        try:
            with open(model, 'r') as f:
                self.json_info = json.load(f)
        except FileNotFoundError:
            try:
                self.json_info = json.loads(model)
                model_as_file = False
            except json.decoder.JSONDecodeError:
                raise Exception("Invalid model argument")

        framework = self.json_info['Framework']
        if 'ModelConfiguration' in self.json_info:
            if isinstance(self.json_info['ModelConfiguration'], str):
                ChildImageClassifier = getattr(importlib.import_module(
                    '{}.{}'.format(framework, self.json_info['ModelConfiguration'])), 'ChildImageClassifier')
            else:
                ChildImageClassifier = getattr(importlib.import_module(
                    '{}.{}'.format(framework, self.json_info['ModelConfiguration']['Name'])), 'ChildImageClassifier')
        else:
            raise Exception("Invalid model configuration")
        os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"
        device = None
        if 'device' in kwargs:
            device = kwargs['device']
            if device == -2:
                device = get_available_device()
        if device is not None:
            if device >= 0:
                try:
                    import torch
                except Exception:
                    raise Exception("PyTorch is not installed. Install it using conda install -c esri deep-learning-essentials")
                torch.cuda.set_device(device)
                arcpy.env.processorType = "GPU"
                arcpy.env.gpuId = str(device)
            else:
                arcpy.env.processorType = "CPU"
        self.child_image_classifier = ChildImageClassifier()
        self.child_image_classifier.initialize(model, model_as_file)
    def getParameterInfo(self):
        required_parameters = [
            {
                'name': 'raster',
                'dataType': 'raster',
                'required': True,
                'displayName': 'Raster',
                'description': 'Input Raster'
            },
            {
                'name': 'model',
                'dataType': 'string',
                'required': True,
                'displayName': 'Input Model Definition (EMD) File',
                'description': 'Input model definition (EMD) JSON file'
            },
            {
                'name': 'device',
                'dataType': 'numeric',
                'required': False,
                'displayName': 'Device ID',
                'description': 'Device ID'
            }
        ]
        return self.child_image_classifier.getParameterInfo(required_parameters)
    def getConfiguration(self, **scalars):
        configuration = self.child_image_classifier.getConfiguration(**scalars)
        if 'DataRange' in self.json_info:
            configuration['dataRange'] = tuple(self.json_info['DataRange'])
        #configuration['inheritProperties'] = 2|4|8
        #configuration['inputMask'] = True
        return configuration
    def updateRasterInfo(self, **kwargs):
        direction = getattr(self.child_image_classifier, "direction", "None")
        tar_nband = self.json_info.get("n_band_c", None)
        if direction == "BtoA":
            kwargs["output_info"]["bandCount"] = int(self.json_info["n_channel_rev"])
        elif tar_nband != None:
             kwargs["output_info"]["bandCount"] = int(tar_nband)
        else:
            kwargs["output_info"]["bandCount"] = int(self.json_info["n_channel"])
        kwargs['output_info']['pixelType'] = 'f4'
        return kwargs
    def updatePixels(self, tlc, shape, props, **pixelBlocks):
        # set pixel values in invalid areas to 0

        # raster_mask = pixelBlocks['raster_mask']
        raster_pixels = pixelBlocks["raster_pixels"]
        # raster_pixels[np.where(raster_mask == 0)] = 0
        pixelBlocks["raster_pixels"] = raster_pixels
        if hasattr(self.child_image_classifier, "updatePixelsSmooth"):
            xx = self.child_image_classifier.updatePixelsSmooth(
                tlc, shape, props, **pixelBlocks
            ).astype(props["pixelType"], copy=False)
            pixelBlocks["output_pixels"] = xx
        else:
            xx = self.child_image_classifier.updatePixels(
                tlc, shape, props, **pixelBlocks
            ).astype(props["pixelType"], copy=False)
            tytx = getattr(
                self.child_image_classifier, "tytx", self.json_info["ImageHeight"]
            )
            chunks, num_rows, num_cols = chunk_it(
                xx.transpose(1, 2, 0), tytx
            )  # self.json_info['ImageHeight'])  # ImageHeight = ImageWidth
            xx = patch_chips(
                crop_flatten(chunks, self.child_image_classifier.padding),
                num_rows,
                num_cols,
            )
            xx = xx.transpose(2, 0, 1)
            pixelBlocks["output_pixels"] = xx
        return pixelBlocks
"""

imagets_classifier_prf = """
import arcpy
import numpy as np
import json
import sys, os, importlib
import math

def get_available_device(max_memory=0.8):
    '''
    select available device based on the memory utilization status of the device
    :param max_memory: the maximum memory utilization ratio that is considered available
    :return: GPU id that is available, -1 means no GPU is available/uses CPU, if GPUtil package is not installed, will
    return 0
    '''
    try:
        import GPUtil
    except ModuleNotFoundError:
        return 0
    GPUs = GPUtil.getGPUs()
    freeMemory = 0
    available = 0
    for GPU in GPUs:
        if GPU.memoryUtil > max_memory:
            continue
        if GPU.memoryFree >= freeMemory:
            freeMemory = GPU.memoryFree
            available = GPU.id
    return available
def chunk_it(image, tile_size):
    s = image.shape
    num_rows = math.ceil(s[0] / tile_size)
    num_cols = math.ceil(s[1] / tile_size)
    r = np.array_split(image, num_rows)
    rows = []
    for x in r:
        x = np.array_split(x, num_cols, axis=1)
        rows.append(x)
    return rows, num_rows, num_cols
def crop_center(img, pad):
    if pad == 0:
        return img
    return img[pad:-pad, pad:-pad, :]
def crop_flatten(chunked, pad):
    imgs = []
    for r, row in enumerate(chunked):
        for c, col in enumerate(row):
            col = crop_center(col, pad)
            imgs.append(col)
    return imgs
def patch_chips(imgs, n_rows, n_cols):
    h_stacks = []
    for i in range(n_rows):
        h_stacks.append(np.hstack(imgs[i * n_cols : n_cols * (i + 1)]))
    return np.vstack(h_stacks)
attribute_table = {
    "displayFieldName": "",
    "fieldAliases": {
        "OID": "OID",
        "Value": "Value",
        "Class": "Class"
    },
    "fields": [
        {"name": "OID", "type": "esriFieldTypeOID", "alias": "OID"},
        {"name": "Value", "type": "esriFieldTypeInteger", "alias": "Value"},
        {"name": "Class", "type": "esriFieldTypeString", "alias": "Class"}
    ],
    "features": [],
}
class ArcGISImageTsClassifier:
    def __init__(self):
        self.name = "ImageTS Classifier"
        self.description = "Image TS classification python raster function to inference a pytorch time-series classifier"

    def initialize(self, **kwargs):
        if "model" not in kwargs:
            return
        model = kwargs["model"]
        model_as_file = True
        try:
            with open(model, "r") as f:
                self.json_info = json.load(f)
        except FileNotFoundError:
            try:
                self.json_info = json.loads(model)
                model_as_file = False
            except json.decoder.JSONDecodeError:
                raise Exception("Invalid model argument")

        framework = self.json_info["Framework"]
        if "ModelConfiguration" in self.json_info:
            if isinstance(self.json_info["ModelConfiguration"], str):
                ChildImageClassifier = getattr(
                    importlib.import_module(
                        "{}.{}".format(framework, self.json_info["ModelConfiguration"])
                    ),
                    "ChildImageClassifier",
                )
            else:
                ChildImageClassifier = getattr(
                    importlib.import_module(
                        "{}.{}".format(
                            framework, self.json_info["ModelConfiguration"]["Name"]
                        )
                    ),
                    "ChildImageClassifier",
                )
        else:
            raise Exception("Invalid model configuration")
        os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"
        device = None
        if "device" in kwargs:
            device = kwargs["device"]
            if device == -2:
                device = get_available_device()
        if device is not None:
            if device >= 0:
                try:
                    import torch
                except Exception:
                    raise Exception(
                        "PyTorch is not installed. Install it using conda install -c esri deep-learning-essentials"
                    )
                torch.cuda.set_device(device)
                arcpy.env.processorType = "GPU"
                arcpy.env.gpuId = str(device)
            else:
                arcpy.env.processorType = "CPU"
        self.child_image_classifier = ChildImageClassifier()
        self.child_image_classifier.initialize(model, model_as_file)
    def getParameterInfo(self):
        required_parameters = [
            {
                "name": "raster",
                "dataType": "raster",
                "required": True,
                "displayName": "Raster",
                "description": "Input Raster",
            },
            {
                "name": "model",
                "dataType": "string",
                "required": True,
                "displayName": "Input Model Definition (EMD) File",
                "description": "Input model definition (EMD) JSON file",
            },
            {
                "name": "device",
                "dataType": "numeric",
                "required": False,
                "displayName": "Device ID",
                "description": "Device ID",
            },
        ]
        return self.child_image_classifier.getParameterInfo(required_parameters)
    def getConfiguration(self, **scalars):
        configuration = self.child_image_classifier.getConfiguration(**scalars)
        if "DataRange" in self.json_info:
            configuration["dataRange"] = tuple(self.json_info["DataRange"])
        configuration["inheritProperties"] = 4 | 8
        configuration["invalidateProperties"] = 2 | 4 | 8
        configuration["inputMask"] = True
        return configuration
    def updateRasterInfo(self, **kwargs):
        kwargs["output_info"]["bandCount"] = 1
        kwargs["output_info"]["pixelType"] = "i4"
        class_info = self.json_info.get("Class_mapping", None)
        num_class_info = self.json_info.get("Num_class_mapping", None)
        if num_class_info == None:
            num_class_info = class_info
        key_vals, num_key_vals = list(class_info.items()), list(num_class_info.items())
        attribute_table["features"] = []
        for i, c in enumerate(class_info):
            attribute_table["features"].append(
                {
                    "attributes": {
                        "OID": i + 1,
                        "Value": num_key_vals[i][1],
                        "Class": key_vals[i][1],
                    }
                }
            )
        kwargs["output_info"]["rasterAttributeTable"] = json.dumps(attribute_table)
        return kwargs
    def updatePixels(self, tlc, shape, props, **pixelBlocks):
        # set pixel values in invalid areas to 0
        raster_mask = pixelBlocks["raster_mask"]
        raster_pixels = pixelBlocks["raster_pixels"]
        raster_pixels[np.where(raster_mask == 0)] = 0
        pixelBlocks["raster_pixels"] = raster_pixels
        xx = self.child_image_classifier.updatePixels(
            tlc, shape, props, **pixelBlocks
        ).astype(props["pixelType"], copy=False)
        tytx = getattr(
            self.child_image_classifier, "tytx", self.json_info["ImageHeight"]
        )
        chunks, num_rows, num_cols = chunk_it(
            xx.transpose(1, 2, 0), tytx
        )  # self.json_info['ImageHeight'])  # ImageHeight = ImageWidth
        xx = patch_chips(
            crop_flatten(chunks, self.child_image_classifier.padding),
            num_rows,
            num_cols,
        )
        xx = xx.transpose(2, 0, 1)
        pixelBlocks["output_pixels"] = xx

        return pixelBlocks
"""

image_captioning_prf = """
import importlib
from importlib import reload, import_module
import json
import os
import sys

import numpy as np
import arcpy


def get_available_device(max_memory=0.8):
    '''
    select available device based on the memory utilization status of the device
    :param max_memory: the maximum memory utilization ratio that is considered available
    :return: GPU id that is available, -1 means no GPU is available/uses CPU, if GPUtil package is not installed, will
    return 0 
    '''
    try:
        import GPUtil
    except ModuleNotFoundError:
        return 0

    GPUs = GPUtil.getGPUs()
    freeMemory = 0
    available = 0
    for GPU in GPUs:
        if GPU.memoryUtil > max_memory:
            continue
        if GPU.memoryFree >= freeMemory:
            freeMemory = GPU.memoryFree
            available = GPU.id

    return available

features = {
    'displayFieldName': '',
    'fieldAliases': {
        'FID': 'FID',
    },
    'geometryType': 'esriGeometryPolygon',
    'fields': [
        {
            'name': 'FID',
            'type': 'esriFieldTypeOID',
            'alias': 'FID'
        }
    ],
    'features': []
}

fields = {
    'fields': [
        {
            'name': 'OID',
            'type': 'esriFieldTypeOID',
            'alias': 'OID'
        },
        {
            'name': 'Shape',
            'type': 'esriFieldTypeGeometry',
            'alias': 'Shape'
        }
    ]
}

class GeometryType:
    Point = 1
    Multipoint = 2
    Polyline = 3
    Polygon = 4


class ArcGISImageCaptioner:
    def __init__(self):
        self.name = 'Image Captioner'
        self.description = 'This python raster function applies deep learning model to caption objects from overlaid imagery'

    def initialize(self, **kwargs):     

        if 'model' not in kwargs:
            return

        # Read esri model definition (emd) file
        model = kwargs['model']
        model_as_file = True

        try:
            with open(model, 'r') as f:
                self.json_info = json.load(f)
        except FileNotFoundError:
            try:
                self.json_info = json.loads(model)
                model_as_file = False
            except json.decoder.JSONDecodeError:
                raise Exception("Invalid model argument")


        framework = self.json_info['Framework']
        if 'ModelConfiguration' in self.json_info:
            modelconfig = self.json_info['ModelConfiguration']
            if isinstance(modelconfig, str):
                if modelconfig not in sys.modules:
                    ChildModelDetector = getattr(import_module(
                        '{}.{}'.format(framework, modelconfig)), 'ChildObjectDetector')
                else:
                    ChildModelDetector = getattr(reload(
                        '{}.{}'.format(framework, modelconfig)), 'ChildObjectDetector')
            else:
                modelconfig = self.json_info['ModelConfiguration']['Name']
                ChildModelDetector = getattr(importlib.import_module(
                    '{}.{}'.format(framework, modelconfig)), 'ChildObjectDetector')
        else:
            raise Exception("Invalid model configuration")

        os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"

        device = None
        if 'device' in kwargs:
            device = kwargs['device']
            if device == -2:
                device = get_available_device()

        if device is not None:
            if device >= 0:
                try:
                    import torch
                except Exception:
                    raise Exception("PyTorch is not installed. Install it using conda install -c esri deep-learning-essentials")
                torch.cuda.set_device(device)
                arcpy.env.processorType = "GPU"
                arcpy.env.gpuId = str(device)
            else:
                arcpy.env.processorType = "CPU"

        self.child_object_detector = ChildModelDetector()
        self.child_object_detector.initialize(model, model_as_file)


    def getParameterInfo(self):

        # PRF needs values of these parameters from gp tool user,
        # either from gp tool UI or emd (a json) file.
        required_parameters = [
            {
                # To support mini batch, it is required that Classify Objects Using Deep Learning geoprocessing Tool
                # passes down a stack of raster tiles to PRF for model inference, the keyword required here is 'rasters'.
                'name': 'rasters',
                'dataType': 'rasters',
                'value': None,
                'required': True,
                'displayName': "Rasters",
                'description': 'The collection of overlapping rasters to objects to be classified'
            },
            {
                'name': 'model',
                'dataType': 'string',
                'required': True,
                'displayName': 'Input Model Definition (EMD) File',
                'description': 'Input model definition (EMD) JSON file'
            },
            {
                'name': 'device',
                'dataType': 'numeric',
                'required': False,
                'displayName': 'Device ID',
                'description': 'Device ID'
            }
        ]

        if 'BatchSize' not in self.json_info:
             required_parameters.append(
                 {
                     'name': 'batch_size',
                     'dataType': 'numeric',
                     'required': False,
                     'value': 4,
                     'displayName': 'Batch Size',
                     'description': 'Batch Size'
                 }
             )

        return self.child_object_detector.getParameterInfo(required_parameters)


    def getConfiguration(self, **scalars):


        # The information PRF returns to the GP tool,
        # the information is either from emd or defined in getConfiguration method.

        configuration = self.child_object_detector.getConfiguration(**scalars)

        if 'DataRange' in self.json_info:
            configuration['dataRange'] = tuple(self.json_info['DataRange'])

        configuration['inheritProperties'] = 2|4|8
        configuration['inputMask'] = True


        return configuration

    def getFields(self):

        fields = {
                'fields': [
                    {
                        'name': 'OID',
                        'type': 'esriFieldTypeOID',
                        'alias': 'OID'
                    },
                    {
                        'name': 'Shape',
                        'type': 'esriFieldTypeGeometry',
                        'alias': 'Shape'
                    }
                ]
            }
        fields['fields'].append(
            {
                'name': 'Caption',
                'type': 'esriFieldTypeString',
                'alias': 'Caption'
            }
        )            
        
        return json.dumps(fields)

    def getGeometryType(self):       
        return GeometryType.Polygon

    def vectorize(self, **pixelBlocks):        
        # set pixel values in invalid areas to 0
        rasters_mask = pixelBlocks['rasters_mask']
        rasters_pixels = pixelBlocks['rasters_pixels']

        for i in range(0, len(rasters_pixels)):
            rasters_pixels[i][np.where(rasters_mask[i] == 0)] = 0

        pixelBlocks['rasters_pixels'] = rasters_pixels

        polygon_list, labels = self.child_object_detector.vectorize(**pixelBlocks)

        features['features'] = []

        features['fieldAliases'].update({
            'Caption':'Caption'
        })

        Labelfield = {
                'name': 'Caption',
                'type': 'esriFieldTypeString',
                'alias': 'Caption'
            }

        if not Labelfield in features['fields']:
            features['fields'].append(Labelfield)        


        for i in range(len(polygon_list)):

            rings = [[]]
            for j in range(len(polygon_list[i])):
                rings[0].append(
                    [
                        polygon_list[i][j][1],
                        polygon_list[i][j][0]
                    ]
                )

            features['features'].append({
                'attributes': {
                    'OID': i + 1,
                    'Caption': labels[i],
                },
                'geometry': {
                    'rings': rings
                }
            })

        return {'output_vectors': json.dumps(features)}
"""
panoptic_segmenter_prf = """
import json
import sys, os, importlib
sys.path.append(os.path.dirname(__file__))

import numpy as np
import math
import arcpy

# utility functions for object detection
def check_centroid_in_center(centroid, start_x, start_y, chip_sz, padding):
    return ((centroid[1] >= (start_y + padding)) and  \
                (centroid[1] <= (start_y + (chip_sz - padding))) and \
                (centroid[0] >= (start_x + padding)) and \
                (centroid[0] <= (start_x + (chip_sz - padding))))

def find_i_j(centroid, n_rows, n_cols, chip_sz, padding, filter_detections):
    for i in range(n_rows):
        for j in range(n_cols):
            start_x = i * chip_sz
            start_y = j * chip_sz

            if (centroid[1] > (start_y)) and (centroid[1] < (start_y + (chip_sz))) and (centroid[0] > (start_x)) and (centroid[0] < (start_x + (chip_sz))):
                in_center = check_centroid_in_center(centroid, start_x, start_y, chip_sz, padding)
                if filter_detections:
                    if in_center:
                        return i, j, in_center
                else:
                    return i, j, in_center
    return None

def get_available_device(max_memory=0.8):
    '''
    select available device based on the memory utilization status of the device
    :param max_memory: the maximum memory utilization ratio that is considered available
    :return: GPU id that is available, -1 means no GPU is available/uses CPU, if GPUtil package is not installed, will
    return 0
    '''
    try:
        import GPUtil
    except ModuleNotFoundError:
        return 0

    GPUs = GPUtil.getGPUs()
    freeMemory = 0
    available = 0
    for GPU in GPUs:
        if GPU.memoryUtil > max_memory:
            continue
        if GPU.memoryFree >= freeMemory:
            freeMemory = GPU.memoryFree
            available = GPU.id

    return available

features = {
    'displayFieldName': '',
    'fieldAliases': {
        'FID': 'FID',
        'Class': 'Class',
        'Confidence': 'Confidence'
    },
    'geometryType': 'esriGeometryPolygon',
    'fields': [
        {
            'name': 'FID',
            'type': 'esriFieldTypeOID',
            'alias': 'FID'
        },
        {
            'name': 'Class',
            'type': 'esriFieldTypeString',
            'alias': 'Class'
        },
        {
            'name': 'Confidence',
            'type': 'esriFieldTypeDouble',
            'alias': 'Confidence'
        }
    ],
    'features': []
}

fields = {
    'fields': [
        {
            'name': 'OID',
            'type': 'esriFieldTypeOID',
            'alias': 'OID'
        },
        {
            'name': 'Class',
            'type': 'esriFieldTypeString',
            'alias': 'Class'
        },
        {
            'name': 'Confidence',
            'type': 'esriFieldTypeDouble',
            'alias': 'Confidence'
        },
        {
            'name': 'Shape',
            'type': 'esriFieldTypeGeometry',
            'alias': 'Shape'
        }
    ]
}

attribute_table = {
    'displayFieldName': '',
    'fieldAliases': {
        'OID': 'OID',
        'Value': 'Value',
        'Class': 'Class',
        'Red': 'Red',
        'Green': 'Green',
        'Blue': 'Blue'
    },
    'fields': [
        {
            'name': 'OID',
            'type': 'esriFieldTypeOID',
            'alias': 'OID'
        },
        {
            'name': 'Value',
            'type': 'esriFieldTypeInteger',
            'alias': 'Value'
        },
        {
            'name': 'Class',
            'type': 'esriFieldTypeString',
            'alias': 'Class'
        },
        {
            'name': 'Red',
            'type': 'esriFieldTypeInteger',
            'alias': 'Red'
        },
        {
            'name': 'Green',
            'type': 'esriFieldTypeInteger',
            'alias': 'Green'
        },
        {
            'name': 'Blue',
            'type': 'esriFieldTypeInteger',
            'alias': 'Blue'
        }
    ],
    'features': []
}

class GeometryType:
    Point = 1
    Multipoint = 2
    Polyline = 3
    Polygon = 4


class ArcGISPanopticSegmenter:
    def __init__(self):
        self.name = 'Panoptic Segmenter'
        self.description = 'This python raster function applies deep learning model to detect object instances in imagery and segment rasters by classes'

    def initialize(self, **kwargs):
        if 'model' not in kwargs:
            return

        model = kwargs['model']
        model_as_file = True
        try:
            with open(model, 'r') as f:
                self.json_info = json.load(f)
        except FileNotFoundError:
            try:
                self.json_info = json.loads(model)
                model_as_file = False
            except json.decoder.JSONDecodeError:
                raise Exception("Invalid model argument")

        # sys.path.append(os.path.dirname(__file__))
        framework = self.json_info['Framework']
        if 'ModelConfiguration' in self.json_info:
            ChildPanopticSegmenter = getattr(importlib.import_module(
                    '{}.{}'.format(framework, self.json_info['ModelConfiguration'])), 'ChildPanopticSegmenter')
        else:
            raise Exception("Invalid model configuration")

        os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"
        device = None
        if 'device' in kwargs:
            device = kwargs['device']
            if device == -2:
                device = get_available_device()

        if device is not None:
            if device >= 0:
                try:
                    import torch
                except Exception:
                    raise Exception("PyTorch is not installed. Install it using conda install -c esri deep-learning-essentials")
                torch.cuda.set_device(device)
                arcpy.env.processorType = "GPU"
                arcpy.env.gpuId = str(device)
            else:
                arcpy.env.processorType = "CPU"

        self.child_panoptic_segmenter = ChildPanopticSegmenter()
        self.child_panoptic_segmenter.initialize(model, model_as_file)

    def getParameterInfo(self):
        required_parameters = [
            {
                'name': 'raster',
                'dataType': 'raster',
                'required': True,
                'displayName': 'Raster',
                'description': 'Input Raster'
            },
            {
                'name': 'model',
                'dataType': 'string',
                'required': True,
                'displayName': 'Input Model Definition (EMD) File',
                'description': 'Input model definition (EMD) JSON file'
            },
            {
                'name': 'device',
                'dataType': 'numeric',
                'required': False,
                'displayName': 'Device ID',
                'description': 'Device ID'
            }
        ]
        return self.child_panoptic_segmenter.getParameterInfo(required_parameters)

    def getConfiguration(self, **scalars):
        configuration = self.child_panoptic_segmenter.getConfiguration(**scalars)
        if 'DataRange' in self.json_info:
            configuration['dataRange'] = tuple(self.json_info['DataRange'])
        configuration['inheritProperties'] = 2|4|8
        configuration['inputMask'] = True
        return configuration

    def getFields(self):
        return json.dumps(fields)

    def getGeometryType(self):
        return GeometryType.Polygon

    def vectorize(self, **pixelBlocks):
        # set pixel values in invalid areas to 0
        raster_mask = pixelBlocks['raster_mask']
        raster_pixels = pixelBlocks['raster_pixels']
        raster_pixels[np.where(raster_mask == 0)] = 0
        pixelBlocks['raster_pixels'] = raster_pixels

        polygon_list, scores, classes = self.child_panoptic_segmenter.vectorize(**pixelBlocks)

        n_rows = int(math.sqrt(self.child_panoptic_segmenter.batch_size))
        n_cols = int(math.sqrt(self.child_panoptic_segmenter.batch_size))
        padding = self.child_panoptic_segmenter.padding
        keep_polygon = []
        keep_scores = []
        keep_classes = []

        for idx, polygon in enumerate(polygon_list):
            centroid = polygon.mean(0)
            quadrant = find_i_j(centroid, n_rows, n_cols, self.json_info['ImageHeight'], padding, self.child_panoptic_segmenter.filter_outer_padding_detections)
            if quadrant is not None:
                i, j, in_center = quadrant
                polygon[:, 0] = polygon[:, 0] - (2*i + 1)*padding
                polygon[:, 1] = polygon[:, 1] - (2*j + 1)*padding
                keep_polygon.append(polygon)
                if not in_center:
                    scores[idx] = (self.child_panoptic_segmenter.thres * 100) + scores[idx] * 0.01
                keep_scores.append(scores[idx])
                keep_classes.append(classes[idx])

        polygon_list =  keep_polygon
        scores = keep_scores
        classes = keep_classes
        features['features'] = []
        for i in range(len(polygon_list)):
            rings = [[]]
            for j in range(polygon_list[i].shape[0]):
                rings[0].append(
                    [
                        polygon_list[i][j][1],
                        polygon_list[i][j][0]
                    ]
                )

            features['features'].append({
                'attributes': {
                    'OID': i + 1,
                    'Class': self.json_info['Classes'][classes[i] - 1]['Name'],
                    'Confidence': scores[i]
                },
                'geometry': {
                    'rings': rings
                }
            })

        return {'output_vectors': json.dumps(features)}

    def updateRasterInfo(self, **kwargs):
        kwargs['output_info']['bandCount'] = 1
        #todo: type is determined by the value range of classes in the json file
        prob_raster = getattr(self.child_panoptic_segmenter,'probability_raster',False)
        if prob_raster:
            kwargs['output_info']['pixelType'] = 'f4' # To ensure that output pixels are in prob range 0 to 1
        else:
            kwargs['output_info']['pixelType'] = 'i4'
        class_info = self.json_info['Classes']
        attribute_table['features'] = []
        for i, c in enumerate(class_info):
            attribute_table['features'].append(
                {
                    'attributes':{
                        'OID':i+1,
                        'Value':c['Value'],
                        'Class':c['Name'],
                        'Red':c['Color'][0],
                        'Green':c['Color'][1],
                        'Blue':c['Color'][2]
                    }
                }
            )
        kwargs['output_info']['rasterAttributeTable'] = json.dumps(attribute_table)

        return kwargs

    def updatePixels(self, tlc, shape, props, **pixelBlocks):
        # set pixel values in invalid areas to 0

        raster_mask = pixelBlocks['raster_mask']
        raster_pixels = pixelBlocks['raster_pixels']
        raster_pixels[np.where(raster_mask == 0)] = 0
        pixelBlocks['raster_pixels'] = raster_pixels

        xx = self.child_panoptic_segmenter.updatePixelsTTA(tlc, shape, props, **pixelBlocks).astype(props['pixelType'], copy=False)
        pixelBlocks['output_pixels'] = xx

        return pixelBlocks

"""

ml_raster_prf = """
import arcpy
import numpy as np
import json
import sys, os, importlib
import math
 
class ArcGISImageClassifier:
    def __init__(self):
        self.name = 'Image Classifier'
        self.description = 'Image classification python raster function to inference a pytorch image classifier'

    def initialize(self, **kwargs):
        if 'model' not in kwargs:
            return

        model = kwargs['model']
        model_as_file = True
        try:
            with open(model, 'r') as f:
                self.json_info = json.load(f)
        except FileNotFoundError:
            try:
                self.json_info = json.loads(model)
                model_as_file = False
            except json.decoder.JSONDecodeError:
                raise Exception("Invalid model argument")
        
        self.class_values = set([row['Value'] for row in self.json_info.get('Classes', [])])
        framework = self.json_info['Framework']
        if 'ModelConfiguration' in self.json_info:
            if isinstance(self.json_info['ModelConfiguration'], str):
                ChildImageClassifier = getattr(importlib.import_module(
                    '{}.{}'.format(framework, self.json_info['ModelConfiguration'])), 'ChildImageClassifier')
            else:
                ChildImageClassifier = getattr(importlib.import_module(
                    '{}.{}'.format(framework, self.json_info['ModelConfiguration']['Name'])), 'ChildImageClassifier')
        else:
            raise Exception("Invalid model configuration")

        os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"
        device = None
        if 'device' in kwargs:
            device = kwargs['device']
            if device == -2:
                device = get_available_device()

        if device is not None:
            if device >= 0:
                try:
                    import torch
                except Exception:
                    raise Exception("PyTorch is not installed. Install it using conda install -c esri deep-learning-essentials")
                torch.cuda.set_device(device)
                arcpy.env.processorType = "GPU"
                arcpy.env.gpuId = str(device)
            else:
                arcpy.env.processorType = "CPU"

        self.child_image_classifier = ChildImageClassifier()
        self.child_image_classifier.initialize(model, model_as_file)

    def getParameterInfo(self):
        required_parameters = [
            {
                'name': 'raster',
                'dataType': 'raster',
                'required': True,
                'displayName': 'Raster',
                'description': 'Input Raster'
            },
            {
                'name': 'model',
                'dataType': 'string',
                'required': True,
                'displayName': 'Input Model Definition (EMD) File',
                'description': 'Input model definition (EMD) JSON file'
            },
            {
                'name': 'device',
                'dataType': 'numeric',
                'required': False,
                'displayName': 'Device ID',
                'description': 'Device ID'
            }
        ]
        params = self.child_image_classifier.getParameterInfo(required_parameters)
        if 0 not in self.class_values and len(self.class_values)>1:
            for param in params:
                if param['name'] == 'predict_background':
                    param['value'] = 'False'
                    break
        return params

    def getConfiguration(self, **scalars):
        configuration = self.child_image_classifier.getConfiguration(**scalars)
        if 'DataRange' in self.json_info:
            configuration['dataRange'] = tuple(self.json_info['DataRange'])
        configuration['inheritProperties'] = 2|4|8
        configuration['inputMask'] = True
        return configuration

    def updateRasterInfo(self, **kwargs):
        kwargs['output_info']['bandCount'] = 1
        #todo: type is determined by the value range of classes in the json file
        prob_raster = getattr(self.child_image_classifier,'probability_raster',False)
        # if prob_raster:
        kwargs['output_info']['pixelType'] = 'f4' # To ensure that output pixels are in prob range 0 to 1
        # else:
        #     kwargs['output_info']['pixelType'] = 'i4'
        # kwargs['output_info']['rasterAttributeTable'] = json.dumps(attribute_table)
        return kwargs


    def updatePixels(self, tlc, shape, props, **pixelBlocks):
        # set pixel values in invalid areas to 0
           
        raster_mask = pixelBlocks['raster_mask']
        raster_pixels = pixelBlocks['raster_pixels']
        raster_pixels[np.where(raster_mask == 0)] = 0
        pixelBlocks['raster_pixels'] = raster_pixels

        xx = self.child_image_classifier.updatePixels(tlc, shape, props, **pixelBlocks).astype(props['pixelType'], copy=False)   

        pixelBlocks['output_pixels'] = xx

        return pixelBlocks
"""



from __future__ import annotations
from arcgis.gis.kubernetes._admin._base import _BaseKube
from arcgis.gis import GIS
from typing import Any, Optional


###########################################################################
class ArchitectureManager(_BaseKube):
    """
    Provides access to the architecture resources defined on the ArcGIS
    Enterprise.
    """

    _gis = None
    _con = None
    _properties = None
    _url = None

    def __init__(self, url: str, gis: GIS):
        super()
        self._url = url
        self._gis = gis
        self._con = gis._con

    # ---------------------------------------------------------------------
    @property
    def development(self) -> dict[str, Any]:
        """
        The development architecture profile is designed for use in
        nonproduction environments, including those for testing and
        evaluation, and requires the least amount of hardware and
        resources. This profile prioritizes replicated pods for publishing
        tools and the private ingress controller, setting replicas for
        both pods at 2. All other pod replicas are set as 1.

        :return: Dict[str, Any]
        """
        url = f"{self._url}/development"
        params = {"f": "json"}
        return self._con.get(url, params)

    # ---------------------------------------------------------------------
    @property
    def standard(self) -> dict[str, Any]:
        """
        The standard-availability architecture profile is designed for use
        in production environments and those wanting to minimize unplanned
        downtime with redundancy across many pods. As a high-availability
        pod, standard-availability provides continued use and availability
        in the even of a failure, and requires less hardware than
        enhanced-availability.

        :return: Dict[str, Any]

        """
        url = f"{self._url}/standard-availability"
        params = {"f": "json"}
        return self._con.get(url, params)

    # ---------------------------------------------------------------------
    @property
    def enhanced(self) -> dict[str, Any]:
        """
        The enhanced-availability architecture profile is designed for use
        in business or mission-critical production environments. This
        profile is designed for the highest level of availability, as it
        includes increased and expanded redundancy across critical pods. As
        a high-availability profile, enhanced-availability provides continued
        use and availability in the event of a failure. However, of the
        available profiles, the hardware requirements are the highest.

        :return: Dict[str, Any]

        """
        url = f"{self._url}/enhanced-availability"
        params = {"f": "json"}
        return self._con.get(url, params)



try:
    import os, sys, json
    import numpy as np
    import torch
    import torch.nn as nn
    import math
    from . import util
    from ._pix2pix import (
        calculate_rectangle_size_from_batch_size,
        get_tile_size,
        tile_to_batch,
        batch_to_tile,
    )

    HAS_TORCH = True
except Exception as e:
    HAS_TORCH = False
import arcgis

from arcgis.learn import PSETAE

try:
    import arcpy
except:
    pass


class ChildImageClassifier:
    def initialize(self, model, model_as_file):
        if not HAS_TORCH:
            raise Exception(
                "Could not find the required deep learning dependencies. Ensure you have installed the required dependent libraries. See https://developers.arcgis.com/python/guide/deep-learning/"
            )

        if arcpy.env.processorType == "GPU" and torch.cuda.is_available():
            self.device = torch.device("cuda")
            arcgis.env._processorType = "GPU"
        else:
            self.device = torch.device("cpu")
            arcgis.env._processorType = "CPU"

        if model_as_file:
            with open(model, "r") as f:
                self.json_info = json.load(f)
        else:
            self.json_info = json.load(model)

        model_path = self.json_info["ModelFile"]
        if model_as_file and not os.path.isabs(model_path):
            model_path = os.path.abspath(
                os.path.join(os.path.dirname(model), model_path)
            )

        self.psetae = PSETAE.from_model(data=None, emd_path=model)
        self._learnmodel = self.psetae
        self.model = self.psetae.learn.model.to(self.device)
        self.model.eval()

    def getParameterInfo(self, required_parameters):
        required_parameters.extend(
            [
                {
                    "name": "padding",
                    "dataType": "numeric",
                    "value": int(self.json_info["ImageHeight"]) // 4,
                    "required": False,
                    "displayName": "Padding",
                    "description": "Padding",
                },
                {
                    "name": "batch_size",
                    "dataType": "numeric",
                    "required": False,
                    "value": 4,
                    "displayName": "Batch Size",
                    "description": "Batch Size",
                },
            ]
        )
        return required_parameters

    def getConfiguration(self, **scalars):
        self.padding = int(
            scalars.get("padding", self.json_info["ImageHeight"] // 4)
        )  ## Default padding Imageheight//4.
        self.batch_size = (
            int(math.sqrt(int(scalars.get("batch_size", 4)))) ** 2
        )  ## Default 4 batch_size

        (
            self.rectangle_height,
            self.rectangle_width,
        ) = calculate_rectangle_size_from_batch_size(self.batch_size)
        ty, tx = get_tile_size(
            self.json_info["ImageHeight"],
            self.json_info["ImageWidth"],
            self.padding,
            self.rectangle_height,
            self.rectangle_width,
        )

        return {"padding": self.padding, "tx": tx, "ty": ty, "fixedTileSize": 1}

    def updatePixels(self, tlc, shape, props, **pixelBlocks):
        input_image = pixelBlocks["raster_pixels"].astype(np.float32)

        batch, batch_height, batch_width = tile_to_batch(
            input_image,
            self.json_info["ImageHeight"],
            self.json_info["ImageWidth"],
            self.padding,
            fixed_tile_size=True,
            batch_height=self.rectangle_height,
            batch_width=self.rectangle_width,
        )

        psetae_ts_prediction = util.pixel_classify_ts_image(
            self.model, batch, self.device, model_info=self.json_info
        )
        psetae_ts_prediction = batch_to_tile(
            psetae_ts_prediction.unsqueeze(dim=1).detach().cpu().numpy(),
            batch_height,
            batch_width,
        )

        return psetae_ts_prediction



reference = {
    "journal": {
        "values": {
            "settings": {
                "appGeocoders": [
                    {
                        "singleLineFieldName": "SingleLine",
                        "name": "Esri World Geocoder",
                        "url": "https://geocode.arcgis.com/arcgis/rest/services/World/GeocodeServer",
                    },
                    {
                        "singleLineFieldName": "SingleLine",
                        "name": "san",
                        "placeholder": "",
                        "url": "https://sampleserver5.arcgisonline.com/arcgis/rest/services/Locators/SanDiego/GeocodeServer",
                    },
                ],
                "layout": {"id": "side"},
            },
            "title": "title",
            "story": {"storage": "WEBAPP", "sections": []},
            "template": {
                "name": "Map Journal",
                "createdWith": "1.14.1",
                "editedWith": "1.14.1",
            },
        }
    }
}

storymap_2 = {
    "root": "n-4xkUEe",
    "nodes": {
        "n-4xkUEe": {
            "type": "story",
            "data": {"storyTheme": "r-vlc4Kp"},
            "config": {"coverDate": "first-published"},
            "children": ["n-aTn8ak", "n-1AItUD", "n-cOeTah"],
        },
        "n-aTn8ak": {
            "type": "storycover",
            "data": {
                "type": "minimal",
                "title": "",
                "summary": "",
                "byline": "",
                "titlePanelPosition": "start",
            },
        },
        "n-1AItUD": {
            "type": "navigation",
            "data": {"links": []},
            "config": {"isHidden": True},
        },
        "n-cOeTah": {"type": "credits"},
    },
    "resources": {
        "r-vlc4Kp": {
            "type": "story-theme",
            "data": {"themeId": "summit", "themeBaseVariableOverrides": {}},
        }
    },
}



# Code from https://github.com/fastai/course-v3/blob/master/nbs/dl2/cyclegan_ws.ipynb
import torch
from fastai.vision import (
    nn,
    Callable,
    List,
    LearnerCallback,
    optim,
    ifnone,
    F,
    flatten_model,
    requires_grad,
    SmoothenValue,
    add_metrics,
)
from .._utils.cyclegan import (
    calculate_activation_statistics,
    calculate_frechet_distance,
)


def convT_norm_relu(
    ch_in: int,
    ch_out: int,
    norm_layer: nn.Module,
    ks: int = 3,
    stride: int = 2,
    bias: bool = True,
):
    return [
        nn.ConvTranspose2d(
            ch_in,
            ch_out,
            kernel_size=ks,
            stride=stride,
            padding=1,
            output_padding=1,
            bias=bias,
        ),
        norm_layer(ch_out),
        nn.ReLU(True),
    ]


def pad_conv_norm_relu(
    ch_in: int,
    ch_out: int,
    pad_mode: str,
    norm_layer: nn.Module,
    ks: int = 3,
    bias: bool = True,
    pad=1,
    stride: int = 1,
    activ: bool = True,
    init: Callable = nn.init.kaiming_normal_,
) -> List[nn.Module]:
    layers = []
    if pad_mode == "reflection":
        layers.append(nn.ReflectionPad2d(pad))
    elif pad_mode == "border":
        layers.append(nn.ReplicationPad2d(pad))
    p = pad if pad_mode == "zeros" else 0
    conv = nn.Conv2d(ch_in, ch_out, kernel_size=ks, padding=p, stride=stride, bias=bias)
    if init:
        init(conv.weight)
        if hasattr(conv, "bias") and hasattr(conv.bias, "data"):
            conv.bias.data.fill_(0.0)
    layers += [conv, norm_layer(ch_out)]
    if activ:
        layers.append(nn.ReLU(inplace=True))
    return layers


class ResnetBlock(nn.Module):
    def __init__(
        self,
        dim: int,
        pad_mode: str = "reflection",
        norm_layer: nn.Module = None,
        dropout: float = 0.0,
        bias: bool = True,
    ):
        super().__init__()
        assert pad_mode in [
            "zeros",
            "reflection",
            "border",
        ], f"padding {pad_mode} not implemented."
        norm_layer = ifnone(norm_layer, nn.InstanceNorm2d)
        layers = pad_conv_norm_relu(dim, dim, pad_mode, norm_layer, bias=bias)
        if dropout != 0:
            layers.append(nn.Dropout(dropout))
        layers += pad_conv_norm_relu(
            dim, dim, pad_mode, norm_layer, bias=bias, activ=False
        )
        self.conv_block = nn.Sequential(*layers)

    def forward(self, x):
        return x + self.conv_block(x)


def resnet_generator(
    ch_in: int,
    ch_out: int,
    n_ftrs: int = 64,
    norm_layer: nn.Module = None,
    dropout: float = 0.0,
    n_blocks: int = 6,
    pad_mode: str = "reflection",
) -> nn.Module:
    norm_layer = ifnone(norm_layer, nn.InstanceNorm2d)
    bias = norm_layer == nn.InstanceNorm2d
    layers = pad_conv_norm_relu(
        ch_in, n_ftrs, "reflection", norm_layer, pad=3, ks=7, bias=bias
    )
    for i in range(2):
        layers += pad_conv_norm_relu(
            n_ftrs, n_ftrs * 2, "zeros", norm_layer, stride=2, bias=bias
        )
        n_ftrs *= 2
    layers += [
        ResnetBlock(n_ftrs, pad_mode, norm_layer, dropout, bias)
        for _ in range(n_blocks)
    ]
    for i in range(2):
        layers += convT_norm_relu(n_ftrs, n_ftrs // 2, norm_layer, bias=bias)
        n_ftrs //= 2
    layers += [
        nn.ReflectionPad2d(3),
        nn.Conv2d(n_ftrs, ch_out, kernel_size=7, padding=0),
        nn.Tanh(),
    ]
    return nn.Sequential(*layers)


def conv_norm_lr(
    ch_in: int,
    ch_out: int,
    norm_layer: nn.Module = None,
    ks: int = 3,
    bias: bool = True,
    pad: int = 1,
    stride: int = 1,
    activ: bool = True,
    slope: float = 0.2,
    init: Callable = nn.init.kaiming_normal_,
) -> List[nn.Module]:
    conv = nn.Conv2d(
        ch_in, ch_out, kernel_size=ks, padding=pad, stride=stride, bias=bias
    )
    if init:
        init(conv.weight)
        if hasattr(conv, "bias") and hasattr(conv.bias, "data"):
            conv.bias.data.fill_(0.0)
    layers = [conv]
    if norm_layer is not None:
        layers.append(norm_layer(ch_out))
    if activ:
        layers.append(nn.LeakyReLU(slope, inplace=True))
    return layers


def discriminator(
    ch_in: int,
    n_ftrs: int = 64,
    n_layers: int = 3,
    norm_layer: nn.Module = None,
    sigmoid: bool = False,
) -> nn.Module:
    norm_layer = ifnone(norm_layer, nn.InstanceNorm2d)
    bias = norm_layer == nn.InstanceNorm2d
    layers = conv_norm_lr(ch_in, n_ftrs, ks=4, stride=2, pad=1)
    for i in range(n_layers - 1):
        new_ftrs = 2 * n_ftrs if i <= 3 else n_ftrs
        layers += conv_norm_lr(
            n_ftrs, new_ftrs, norm_layer, ks=4, stride=2, pad=1, bias=bias
        )
        n_ftrs = new_ftrs
    new_ftrs = 2 * n_ftrs if n_layers <= 3 else n_ftrs
    layers += conv_norm_lr(
        n_ftrs, new_ftrs, norm_layer, ks=4, stride=1, pad=1, bias=bias
    )
    layers.append(nn.Conv2d(new_ftrs, 1, kernel_size=4, stride=1, padding=1))
    if sigmoid:
        layers.append(nn.Sigmoid())
    return nn.Sequential(*layers)


class CycleGAN(nn.Module):
    def __init__(
        self,
        ch_in: int,
        ch_out: int,
        n_features: int = 64,
        disc_layers: int = 3,
        gen_blocks: int = 6,
        lsgan: bool = True,
        drop: float = 0.0,
        norm_layer: nn.Module = None,
    ):
        super().__init__()
        self.D_A = discriminator(
            ch_in, n_features, disc_layers, norm_layer, sigmoid=not lsgan
        )
        self.D_B = discriminator(
            ch_in, n_features, disc_layers, norm_layer, sigmoid=not lsgan
        )
        self.G_A = resnet_generator(
            ch_in, ch_out, n_features, norm_layer, drop, gen_blocks
        )
        self.G_B = resnet_generator(
            ch_in, ch_out, n_features, norm_layer, drop, gen_blocks
        )
        # G_A: takes real input B and generates fake input A
        # G_B: takes real input A and generates fake input B
        # D_A: trained to make the difference between real input A and fake input A
        # D_B: trained to make the difference between real input B and fake input B
        self.arcgis_results = False

    def forward(self, real_A, real_B):
        fake_A, fake_B = self.G_A(real_B), self.G_B(real_A)
        if self.arcgis_results:
            return torch.cat([fake_A[:, None], fake_B[:, None]], 1)
        idt_A, idt_B = (
            self.G_A(real_A),
            self.G_B(real_B),
        )  # Needed for the identity loss during training.
        return [fake_A, fake_B, idt_A, idt_B]


class AdaptiveLoss(nn.Module):
    def __init__(self, crit):
        super().__init__()
        self.crit = crit

    def forward(self, output, target: bool, **kwargs):
        targ = (
            output.new_ones(*output.size())
            if target
            else output.new_zeros(*output.size())
        )
        return self.crit(output, targ, **kwargs)


class CycleGanLoss(nn.Module):
    def __init__(
        self,
        cgan: nn.Module,
        lambda_A: float = 10.0,
        lambda_B: float = 10,
        lambda_idt: float = 0.5,
        lsgan: bool = True,
    ):
        super().__init__()
        self.cgan, self.l_A, self.l_B, self.l_idt = cgan, lambda_A, lambda_B, lambda_idt
        self.crit = AdaptiveLoss(F.mse_loss if lsgan else F.binary_cross_entropy)

    def set_input(self, input):
        self.real_A, self.real_B = input

    def forward(self, output, target):
        fake_A, fake_B, idt_A, idt_B = output
        # Generators should return identity on the datasets they try to convert to
        self.id_loss = self.l_idt * (
            self.l_A * F.l1_loss(idt_A, self.real_A)
            + self.l_B * F.l1_loss(idt_B, self.real_B)
        )
        # Generators are trained to trick the discriminators so the following should be ones
        self.gen_loss = self.crit(self.cgan.D_A(fake_A), True) + self.crit(
            self.cgan.D_B(fake_B), True
        )
        # Cycle loss
        self.cyc_loss = self.l_A * F.l1_loss(self.cgan.G_A(fake_B), self.real_A)
        self.cyc_loss += self.l_B * F.l1_loss(self.cgan.G_B(fake_A), self.real_B)
        return self.id_loss + self.gen_loss + self.cyc_loss


class CycleGANTrainer(LearnerCallback):
    _order = -20  # Need to run before the Recorder

    def _set_trainable(self, D_A=False, D_B=False):
        gen = (not D_A) and (not D_B)
        ## Removing gradient flags which are causing incorrect
        ## results on training the model second time.
        # requires_grad(self.learn.model.G_A, gen)
        # requires_grad(self.learn.model.G_B, gen)
        # requires_grad(self.learn.model.D_A, D_A)
        # requires_grad(self.learn.model.D_B, D_B)
        if not gen:
            self.opt_D_A.lr, self.opt_D_A.mom = self.learn.opt.lr, self.learn.opt.mom
            self.opt_D_A.wd, self.opt_D_A.beta = self.learn.opt.wd, self.learn.opt.beta
            self.opt_D_B.lr, self.opt_D_B.mom = self.learn.opt.lr, self.learn.opt.mom
            self.opt_D_B.wd, self.opt_D_B.beta = self.learn.opt.wd, self.learn.opt.beta

    def on_train_begin(self, **kwargs):
        self.G_A, self.G_B = self.learn.model.G_A, self.learn.model.G_B
        self.D_A, self.D_B = self.learn.model.D_A, self.learn.model.D_B
        self.crit = self.learn.loss_func.crit
        if not getattr(self, "opt_G", None):
            self.opt_G = self.learn.opt.new(
                [nn.Sequential(*flatten_model(self.G_A), *flatten_model(self.G_B))]
            )
        else:
            self.opt_G.lr, self.opt_G.wd = self.opt.lr, self.opt.wd
            self.opt_G.mom, self.opt_G.beta = self.opt.mom, self.opt.beta
        if not getattr(self, "opt_D_A", None):
            self.opt_D_A = self.learn.opt.new([nn.Sequential(*flatten_model(self.D_A))])
        if not getattr(self, "opt_D_B", None):
            self.opt_D_B = self.learn.opt.new([nn.Sequential(*flatten_model(self.D_B))])
        self.learn.opt.opt = self.opt_G.opt
        self._set_trainable()
        self.id_smter, self.gen_smter, self.cyc_smter = (
            SmoothenValue(0.98),
            SmoothenValue(0.98),
            SmoothenValue(0.98),
        )
        self.da_smter, self.db_smter = SmoothenValue(0.98), SmoothenValue(0.98)
        self.recorder.add_metric_names(
            ["id_loss", "gen_loss", "cyc_loss", "D_A_loss", "D_B_loss"]
        )

    def on_batch_begin(self, last_input, **kwargs):
        self.learn.loss_func.set_input(last_input)

    def on_backward_begin(self, **kwargs):
        self.id_smter.add_value(self.loss_func.id_loss.detach().cpu())
        self.gen_smter.add_value(self.loss_func.gen_loss.detach().cpu())
        self.cyc_smter.add_value(self.loss_func.cyc_loss.detach().cpu())

    def on_batch_end(self, last_input, last_output, **kwargs):
        self.G_A.zero_grad()
        self.G_B.zero_grad()
        fake_A, fake_B = last_output[0].detach(), last_output[1].detach()
        real_A, real_B = last_input
        self._set_trainable(D_A=True)
        self.D_A.zero_grad()
        loss_D_A = 0.5 * (
            self.crit(self.D_A(real_A), True) + self.crit(self.D_A(fake_A), False)
        )
        self.da_smter.add_value(loss_D_A.detach().cpu())
        if self.learn.model.training == True:
            loss_D_A.backward()
        # loss_D_A.backward()
        self.opt_D_A.step()
        self._set_trainable(D_B=True)
        self.D_B.zero_grad()
        loss_D_B = 0.5 * (
            self.crit(self.D_B(real_B), True) + self.crit(self.D_B(fake_B), False)
        )
        self.db_smter.add_value(loss_D_B.detach().cpu())
        if self.learn.model.training == True:
            loss_D_B.backward()
        # loss_D_B.backward()
        self.opt_D_B.step()
        self._set_trainable()

    def on_epoch_end(self, last_metrics, **kwargs):
        return add_metrics(
            last_metrics,
            [
                s.smooth
                for s in [
                    self.id_smter,
                    self.gen_smter,
                    self.cyc_smter,
                    self.da_smter,
                    self.db_smter,
                ]
            ],
        )


def compute_fid_metric(model, data, a_or_b):
    if a_or_b == "a":
        idx = 0
    else:
        idx = 1
    input_idx = []
    pred_idx = []

    for input, target in data.valid_dl:
        input_idx.append(input[idx] / 2 + 0.5)
        pred = model.learn.pred_batch(batch=(input, target))
        pred_idx.append(pred[idx] / 2 + 0.5)

    data_len = len(data.valid_ds)
    batch_size = data.batch_size

    m1, s1 = calculate_activation_statistics(batch_size, data_len, input_idx)
    m2, s2 = calculate_activation_statistics(batch_size, data_len, pred_idx)

    fid_value = calculate_frechet_distance(m1, s1, m2, s2)

    return fid_value



import sys
import logging
from typing import Dict, Any, Tuple
from .tools._util import check_module_exists

__log__ = logging.getLogger()

if sys.platform == "win32" and check_module_exists("certifi_win32"):  # pragma: no cover
    # when on Windows, append to the certifi
    # the users trusted certificate store
    # when certifi_win32 is present.
    try:
        import certifi_win32

        certifi_win32.wincerts.verify_combined_pem()
        certifi_win32.wincerts.where()
    except ImportError:
        pass
elif check_module_exists("truststore"):  # pragma: no cover
    try:
        import truststore

        truststore.inject_into_ssl()
    except ImportError as ie:
        __log__.warning(f"truststore raised a warning: {ie}")
    except Exception as e:
        __log__.warning(f"truststore raised a warning: {e}")


from requests.sessions import Session
from requests.adapters import HTTPAdapter


from urllib3 import Retry
from urllib3 import __version__ as __URLLIB3VERSION__

__URLLIB3VERSION__ = [
    int(i) if i.isdigit() else i for i in __URLLIB3VERSION__.split(".")
]

from ._version import __version__

from ._auth import (
    EsriPKIAuth,
    EsriWindowsAuth,
    EsriKerberosAuth,
)

from ._auth._winauth import HAS_KERBEROS
from ._auth._negotiate import HAS_GSSAPI

from requests_toolbelt.adapters.host_header_ssl import HostHeaderSSLAdapter

from .tools import LazyLoader

urllib3 = LazyLoader("urllib3")


###########################################################################
class EsriSession:
    """
    The `EsriSession` class is designed to simplify access to the Esri WebGIS
    environments without the additional components of the `arcgis` or `arcpy`
    modules.  It is designed to allow users to connect and manage all HTTP calls
    themselves with no hand holding.

    Security is handled through the `requests` authentication model. Leveraging
    authentication handlers and primarily connecting to 10.8.1+ enterprise components.

    The supported Authentication Schemes are:

        1. Username/Password
        2. Oauth 2
        3. IWA/NTLM
        4. Basic
        5. Digest
        6. PKI
        7. API Keys
        8. User Provided Tokens
        9. Kerberos

    Anyone can extend the authentication handlers by creating a class and inheriting
    from `requests.auth.AuthBase`.

    The `EsriSession` authentication supports authentication chaining.  Meaning
    users can stack authentication methods.

        `auth1 + auth2 + auth3`

    It is recommended that you do not stack unneeded authenicators because they
    can caused unintended failures.

    ==================     ====================================================================
    **Parameter**           **Description**
    ------------------     --------------------------------------------------------------------
    auth                   Optional AuthBase. This is a security handler that performs some sort
                           of security check.
    ------------------     --------------------------------------------------------------------
    cert                   Optional Tuple. The client side certificate as a tuple or string. It
                           should be noted that `EsriSession` does not support encrypted private
                           keys.
    ------------------     --------------------------------------------------------------------
    verify_cert            Optional Bool. When `False` all SSL certificate errors are ignored.
                           The default is `True`.
    ------------------     --------------------------------------------------------------------
    allow_redirects        Optional Bool. When `False` if the URL redirects a user, an error
                           will be raised.  The default is `True`
    ------------------     --------------------------------------------------------------------
    headers                Optional Dict. An additional set of key/value(s) to append to any
                           request's header.
    ------------------     --------------------------------------------------------------------
    referer                Optional Str. The `referer` header value.  The default is `http`. This
                           is mainly used with `token` security.
    ==================     ====================================================================


    **Optional Arguments**

    ==================     ====================================================================
    **Parameter**           **Description**
    ------------------     --------------------------------------------------------------------
    trust_env              Optional Bool. The default is `True`. If `False` proxies will cause
                           an error to be raised if set by **.netrc** files.
    ------------------     --------------------------------------------------------------------
    stream                 Optional Bool.  To enable handling streaming responses, set stream to
                           True and iterate over the response with `iter_lines`. The default is
                           `False`.
    ------------------     --------------------------------------------------------------------
    check_hostname         Optional Bool. When connecting to a side via IP Address with an SSL
                           certificate, the hostname will not match.  This allows a user to
                           specify the hostname in the headers parameters and ignore hostname
                           errors.
    ------------------     --------------------------------------------------------------------
    retries                Optional Int. The max number of tries to retry a service for 50x errors.
    ------------------     --------------------------------------------------------------------
    backoff_factor         Optional float. The amount of time in seconds to wait between retries.
    ------------------     --------------------------------------------------------------------
    status_to_retry        Optional Tuple. The status codes to run retries on.  The default is
                           (413, 429, 503, 500, 502, 504).
    ------------------     --------------------------------------------------------------------
    method_whitelist       Optional List.  When `retries` is specified, the user can specifiy what methods are retried.
                           The default is `'POST', 'DELETE', 'GET', 'HEAD', 'OPTIONS', 'PUT', 'TRACE'`
    ------------------     --------------------------------------------------------------------
    proxies                Optional Dict. A key/value mapping where the keys are the transfer protocol and the value is the <url>:<port>.

                           **example**

                           ```python
                           proxies = {"http" : 127.0.0.1:8080, "https" : 127.0.0.1:8081}
                           session = EsriSession(proxies=proxies)
                           ```

    ==================     ====================================================================


    """

    _session = None
    _verify = None
    _baseurl = None  # if partial url given, try the base url
    _referer = None
    allow_redirects = None

    # ----------------------------------------------------------------------
    def __init__(
        self,
        auth: "AuthBase" = None,
        cert: Tuple[str] = None,
        verify_cert: bool = True,
        allow_redirects: bool = True,
        headers: Dict[str, Any] = None,
        referer="http",
        **kwargs,
    ) -> "EsriSession":
        super()

        self._session = Session()
        self._session.stream = kwargs.pop("stream", False)
        check_hostname = kwargs.get("check_hostname", True)
        self._session.trust_env = kwargs.pop("trust_env", True)
        self._prevent_keep_alive = kwargs.pop("keep_alive", False)
        if check_hostname == False:
            self.mount("https://", HostHeaderSSLAdapter())
        self._session.cert = cert
        self._cert = cert
        self.allow_redirects = allow_redirects
        self.verify_cert = verify_cert
        self._useragent = f"EsriSession/{__version__}"
        self._session.headers["User-Agent"] = self._useragent
        if referer is None:
            referer = ""
        self._referer = referer
        if isinstance(headers, dict):
            self.update_headers(headers)
        if not "referer" in self._session.headers:
            self._session.headers["referer"] = self._referer

        if auth and cert is None:
            self._session.auth = auth

        elif auth and cert:
            self.auth = EsriPKIAuth(
                cert=cert,
                referer=referer,
                verify_cert=verify_cert,
                auth=auth,
                session=self,
            )
        elif auth is None and cert:
            self.auth = EsriPKIAuth(
                cert=cert,
                referer=referer,
                verify_cert=verify_cert,
                session=self,
            )
        elif sys.platform == "win32" and HAS_GSSAPI:  # Default Case Load IWA/WinAuth
            self.auth = EsriWindowsAuth(referer=referer, verify_cert=verify_cert)
        elif HAS_KERBEROS:
            self.auth = EsriKerberosAuth(referer=self._referer, verify_cert=verify_cert)

        proxies = kwargs.get("proxies", None)
        if proxies:
            self.proxies = proxies

        if "retries" in kwargs and kwargs.get("retries"):
            if __URLLIB3VERSION__[0] <= 1:
                r = Retry(
                    total=kwargs.get("retries", 5),
                    read=kwargs.get("retries", 5),
                    connect=kwargs.get("retries", 5),
                    status_forcelist=kwargs.get(
                        "status_to_retry", (413, 429, 503, 500, 502, 504)
                    ),
                    method_whitelist=kwargs.get(
                        "method_whitelist",
                        frozenset(
                            [
                                "POST",
                                "DELETE",
                                "GET",
                                "HEAD",
                                "OPTIONS",
                                "PUT",
                                "TRACE",
                            ]
                        ),
                    ),
                )
            else:
                r = Retry(
                    total=kwargs.get("retries", 5),
                    read=kwargs.get("retries", 5),
                    connect=kwargs.get("retries", 5),
                    status_forcelist=kwargs.get(
                        "status_to_retry", (413, 429, 503, 500, 502, 504)
                    ),
                    allowed_methods=kwargs.get(
                        "method_whitelist",
                        frozenset(
                            [
                                "POST",
                                "DELETE",
                                "GET",
                                "HEAD",
                                "OPTIONS",
                                "PUT",
                                "TRACE",
                            ]
                        ),
                    ),
                )
            adapter = HTTPAdapter(max_retries=r)
            self._session.mount("http://", adapter)
            self._session.mount("https://", adapter)

    # ----------------------------------------------------------------------
    def close(self):
        """Closes all adapters and as such the session"""
        self._session.close()

    # ----------------------------------------------------------------------
    def __enter__(self) -> "EsriSession":
        return self

    # ----------------------------------------------------------------------
    def __exit__(self, *args):
        self._session.close()

    # ----------------------------------------------------------------------
    @property
    def stream(self) -> bool:
        """Gets/Sets the stream property for the current session object"""
        return self._session.stream

    # ----------------------------------------------------------------------
    @stream.setter
    def stream(self, stream: bool):
        """Gets/Sets the stream property for the current session object"""
        if isinstance(stream, bool):
            self._session.stream = stream

    # ----------------------------------------------------------------------
    @property
    def headers(self) -> Dict[str, Any]:
        """Gets/Sets the headers from the current session object"""
        return self._session.headers

    # ----------------------------------------------------------------------
    @headers.setter
    def headers(self, values: Dict[str, Any]):
        """Gets/Sets the headers from the current session object"""
        if isinstance(values, dict):
            from requests.utils import CaseInsensitiveDict

            values = CaseInsensitiveDict(values)
            self._session.headers = values

    # ----------------------------------------------------------------------
    def update_headers(self, values: Dict[str, Any]) -> bool:
        """Performs an update call on the headers"""
        try:
            self._session.headers.update(values)
            return True
        except:
            return False

    # ----------------------------------------------------------------------
    @property
    def referer(self) -> str:
        """Gets/Sets the referer"""
        try:
            return self._session.headers["referer"]
        except:
            return None

    # ----------------------------------------------------------------------
    @referer.setter
    def referer(self, value: str):
        """Gets/Sets the referer"""
        self._session.headers["referer"] = value

    # ----------------------------------------------------------------------
    def __str__(self) -> str:
        return f"<ArcGIS Session {__version__}>"

    # ----------------------------------------------------------------------
    def __repr__(self) -> str:
        return f"<ArcGIS Session {__version__}>"

    # ----------------------------------------------------------------------
    @property
    def verify_cert(self) -> bool:
        """
        Get/Set property that allows for the verification of SSL certificates

        :returns: bool
        """
        return self._session.verify

    # ----------------------------------------------------------------------
    @verify_cert.setter
    def verify_cert(self, value: bool):
        if isinstance(value, bool) and value != self._session.verify:
            self._session.verify = value

    # ----------------------------------------------------------------------
    def mount(self, prefix: str, adapter: "HTTPAdatper"):
        """
        Registers a connection adapter to a prefix.

        Adapters are sorted in descending order by prefix length.
        """
        return self._session.mount(prefix, adapter)

    # ----------------------------------------------------------------------
    @property
    def adapters(self) -> Dict[str, str]:
        """
        Returns an dictionary of mounted adapters.

        :return: dict
        """
        return self._session.adapters

    # ----------------------------------------------------------------------
    @property
    def auth(self) -> "AuthBase":
        """Get/Set the Authentication Handler for the Session"""
        return self._session.auth

    # ----------------------------------------------------------------------
    @auth.setter
    def auth(self, value: "AuthBase"):
        """Get/Set the Authentication Handler for the Session"""
        self._session.auth = value

    # ----------------------------------------------------------------------
    @property
    def proxies(self) -> Dict[str, str]:
        """
        Dictionary mapping protocol or protocol and host to the URL of the proxy.
        (e.g. {'http': 'foo.bar:3128', 'http://host.name': 'foo.bar:4012'}) to
        be used on each :class:`Request <Request>`.

        :return: dict
        """
        return self._session.proxies

    # ----------------------------------------------------------------------
    @proxies.setter
    def proxies(self, value: Dict[str, str]) -> None:
        """
        Dictionary mapping protocol or protocol and host to the URL of the proxy.
        (e.g. {'http': 'foo.bar:3128', 'http://host.name': 'foo.bar:4012'}) to
        be used on each :class:`Request <Request>`.

        :return: dict
        """
        if isinstance(value, dict) and self._session.proxies:
            self._session.proxies.update(value)
        elif isinstance(value, dict) and not self._session.proxies:
            self._session.proxies = value
        elif value is None:
            self._session.proxies = {}
        else:
            raise ValueError("Proxy must be of type dictionary.")

    @property
    def cert(self) -> Tuple[str]:
        """
        Get/Set the users certificate as a (private, public) keys.

        :return: Tuple[str]
        """
        return self._cert or self._session.cert

    @cert.setter
    def cert(self, cert: Tuple[str]):
        """
        Get/Set the users certificate as a (private, public) keys.

        :return: Tuple[str]
        """
        if cert is None:
            self._cert = None
            self._session.cert = None
            if isinstance(self._session.auth, EsriPKIAuth):
                self._session.auth = None
        else:
            self._cert = cert
            self._session.cert = cert
            self._session.auth = EsriPKIAuth(
                cert=cert,
                referer=self._referer,
                verify_cert=self.verify_cert,
            )

    # ----------------------------------------------------------------------
    def get(self, url, **kwargs) -> "requests.Response":
        r"""Sends a GET request. Returns :class:`Response` object.

        :param url: URL for the new :class:`Request` object.
        :param \*\*kwargs: Optional arguments that ``request`` takes.
        :rtype: requests.Response
        """
        if "allow_redirects" in kwargs:
            redirects = kwargs.pop("allow_redirects")
        else:
            redirects = self.allow_redirects
        if "proxies" in kwargs:
            proxies = kwargs.pop("proxies")
        else:
            proxies = self.proxies
        return self._session.get(
            url, allow_redirects=redirects, proxies=proxies, **kwargs
        )

    # ----------------------------------------------------------------------
    def options(self, url, **kwargs) -> "requests.Response":
        r"""Sends a OPTIONS request. Returns :class:`Response` object.

        :param url: URL for the new :class:`Request` object.
        :param \*\*kwargs: Optional arguments that ``request`` takes.
        :rtype: requests.Response
        """
        if "proxies" in kwargs:
            proxies = kwargs.pop("proxies")
        else:
            proxies = self.proxies
        return self._session.options(url, proxies=proxies, **kwargs)

    # ----------------------------------------------------------------------
    def head(self, url, **kwargs) -> "requests.Response":
        r"""Sends a HEAD request. Returns :class:`Response` object.

        :param url: URL for the new :class:`Request` object.
        :param \*\*kwargs: Optional arguments that ``request`` takes.
        :rtype: requests.Response
        """
        if "proxies" in kwargs:
            proxies = kwargs.pop("proxies")
        else:
            proxies = self.proxies
        return self._session.head(url, proxies=proxies, **kwargs)

    # ----------------------------------------------------------------------
    def post(self, url, data=None, json=None, **kwargs) -> "requests.Response":
        r"""Sends a POST request. Returns :class:`Response` object.

        :param url: URL for the new :class:`Request` object.
        :param data: (optional) Dictionary, list of tuples, bytes, or file-like
            object to send in the body of the :class:`Request`.
        :param json: (optional) json to send in the body of the :class:`Request`.
        :param \*\*kwargs: Optional arguments that ``request`` takes.
        :rtype: requests.Response
        """
        if "proxies" in kwargs:
            proxies = kwargs.pop("proxies")
        else:
            proxies = self.proxies
        if "allow_redirects" in kwargs:
            redirects = kwargs.pop("allow_redirects")
        else:
            redirects = self.allow_redirects
        return self._session.post(
            url,
            data=data,
            json=json,
            allow_redirects=redirects,
            proxies=proxies,
            **kwargs,
        )

    # ----------------------------------------------------------------------
    def put(self, url, data=None, **kwargs) -> "requests.Response":
        r"""Sends a PUT request. Returns :class:`Response` object.

        :param url: URL for the new :class:`Request` object.
        :param data: (optional) Dictionary, list of tuples, bytes, or file-like
            object to send in the body of the :class:`Request`.
        :param \*\*kwargs: Optional arguments that ``request`` takes.
        :rtype: requests.Response
        """
        if "proxies" in kwargs:
            proxies = kwargs.pop("proxies")
        else:
            proxies = self.proxies
        return self._session.put(url, data=data, proxies=proxies, **kwargs)

    # ----------------------------------------------------------------------
    def patch(self, url, data=None, **kwargs) -> "requests.Response":
        r"""Sends a PATCH request. Returns :class:`Response` object.

        :param url: URL for the new :class:`Request` object.
        :param data: (optional) Dictionary, list of tuples, bytes, or file-like
            object to send in the body of the :class:`Request`.
        :param \*\*kwargs: Optional arguments that ``request`` takes.
        :rtype: requests.Response
        """
        if "proxies" in kwargs:
            proxies = kwargs.pop("proxies")
        else:
            proxies = self.proxies
        return self._session.patch(url, data=data, proxies=proxies, **kwargs)

    # ----------------------------------------------------------------------
    def delete(self, url, **kwargs) -> "requests.Response":
        r"""Sends a DELETE request. Returns :class:`Response` object.

        :param url: URL for the new :class:`Request` object.
        :param \*\*kwargs: Optional arguments that ``request`` takes.
        :rtype: requests.Response
        """
        if "proxies" in kwargs:
            proxies = kwargs.pop("proxies")
        else:
            proxies = self.proxies
        return self._session.delete(url, proxies=proxies, **kwargs)



from arcgis.features import Feature, FeatureSet
from arcgis.features import FeatureLayer, Table
from arcgis.geometry import Geometry
import numpy as np
import pandas as pd


# --------------------------------------------------------------------------
def _chunks(l, n):
    """yield successive n-sized chunks from l."""
    for i in range(0, len(l), n):
        yield l[i : i + n]


if [float(i) for i in pd.__version__.split(".")] < [1, 0, 0]:
    _look_up_types = {
        "esriFieldTypeSmallInteger": np.int32,
        "esriFieldTypeInteger": np.int32,
        "esriFieldTypeSingle": float,
        "esriFieldTypeDouble": float,
        "esriFieldTypeFloat": float,
        "esriFieldTypeString": str,
        "esriFieldTypeDate": pd.datetime,
        "esriFieldTypeOID": np.int64,
        "esriFieldTypeGeometry": object,
        "esriFieldTypeBlob": object,
        "esriFieldTypeRaster": object,
        "esriFieldTypeGUID": str,
        "esriFieldTypeGlobalID": str,
        "esriFieldTypeXML": object,
        "esriFieldTypeBigInteger": np.int64,
        "esriFieldTypeTimeOnly": str,
        "esriFieldTypeDateOnly": pd.datetime,
        "esriFieldTypeTimestampOffset": str,
    }
else:
    from datetime import datetime as _datetime

    _look_up_types = {
        "esriFieldTypeSmallInteger": pd.Int32Dtype(),
        "esriFieldTypeInteger": pd.Int32Dtype(),
        "esriFieldTypeSingle": pd.Float64Dtype(),
        "esriFieldTypeDouble": pd.Float64Dtype(),
        "esriFieldTypeFloat": pd.Float64Dtype(),
        "esriFieldTypeString": pd.StringDtype(),
        "esriFieldTypeDate": "<M8[us]",
        "esriFieldTypeOID": pd.Int64Dtype(),
        "esriFieldTypeGeometry": object,
        "esriFieldTypeBlob": object,
        "esriFieldTypeRaster": object,
        "esriFieldTypeGUID": pd.StringDtype(),
        "esriFieldTypeGlobalID": pd.StringDtype(),
        "esriFieldTypeXML": object,
        "esriFieldTypeBigInteger": pd.Int64Dtype(),  #  added 11.2
        "esriFieldTypeTimeOnly": pd.StringDtype(),  #  added 11.2
        "esriFieldTypeDateOnly": "<M8[us]",  #  added 11.2
        "esriFieldTypeTimestampOffset": "<M8[us]",  #  added 11.2
    }


# --------------------------------------------------------------------------
def to_featureset(df):
    """converts a pd.DataFrame to a FeatureSet Object"""
    if hasattr(df, "spatial"):
        fs = df.spatial.__feature_set__
        return FeatureSet.from_dict(fs)
    return None


# --------------------------------------------------------------------------
def from_featureset(fset, sr=None):
    """

    Converts a FeatureSet to a pd.DataFrame

    ===============    ==============================================
    Arguments          Description
    ---------------    ----------------------------------------------
    fset               Required FeatureSet.  FeatureSet object.
    ===============    ==============================================

    return Panda's DataFrame

    """
    if isinstance(fset, FeatureSet):
        rows = []
        sr = fset.spatial_reference
        try:
            gt = fset.geometry_type.replace("esriGeometry", "")
        except:
            gt = None
        cols = [fld["name"] for fld in fset.fields]
        dt_fields = [
            fld["name"]
            for fld in fset.fields
            if ("type" in fld and fld["type"] == "esriFieldTypeDate")
            or ("fieldType" in fld and fld["fieldType"] == "esriFieldTypeDate")
        ]

        pandas_dtypes = {}
        for fld in fset.fields:
            if fld["name"].lower() != "shape":
                if "type" in fld:
                    pandas_dtypes[fld["name"]] = _look_up_types[fld["type"]]
                elif "fieldType" in fld:
                    pandas_dtypes[fld["name"]] = _look_up_types[fld["fieldType"]]
                else:
                    pandas_dtypes[fld["name"]] = "O"

        if sr is None:
            sr = {"wkid": 4326}
        for feat in fset.features:
            a = feat.attributes
            if feat.geometry is not None:
                g = feat.geometry
                g["spatialReference"] = sr
                a["SHAPE"] = Geometry(g)
            rows.append(a)
            del a, feat
        from arcgis.features import GeoAccessor, GeoSeriesAccessor

        if len(rows) > 0:
            for row in rows:
                if len(set(row.keys()) - set(cols)) > 0:
                    # break at first occurrence
                    cols = list(row.keys())
                    break

        df = pd.DataFrame(data=rows, columns=cols)

        for fld in dt_fields:
            try:
                df[fld] = pd.to_datetime(df[fld] / 1000, unit="s")
            except:
                df[fld] = pd.to_datetime(
                    df[fld],
                )

        if "SHAPE" in df.columns:
            # replace the float NaN with None, otherwise error occurs
            df["SHAPE"].replace({np.nan: None}, inplace=True)
            df.spatial.set_geometry("SHAPE")
            df.spatial.sr = sr
            for i in range(len(df)):
                shape = df.loc[i]["SHAPE"]
                # Check if NaN by comparing to self.
                if shape != shape:
                    df.iat[i, df.columns.get_loc("SHAPE")] = None
        if pandas_dtypes:
            try:
                df = df.astype(pandas_dtypes)
                df = df.convert_dtypes()
                return df
            except:
                return df.convert_dtypes()
        return df
    else:
        return None


# --------------------------------------------------------------------------
def from_layer(layer, query="1=1"):
    """
    Converts a Feature Service Layer to a Pandas' DataFrame

    Parameters:
     :layer: FeatureLayer or Table object.  If the object is a FeatureLayer
      the function will return a Spatial DataFrame, if the object is of
      type Table, the function will return a Pandas' DataFrame


    Usage:

    >>> from arcgis.features import FeatureLayer
    >>> mylayer = FeatureLayer(("https://sampleserver6.arcgisonline.com/arcgis/rest"
                        "/services/CommercialDamageAssessment/FeatureServer/0"))
    >>> df = from_layer(mylayer)
    >>> print(df.head())

    """
    if not layer.filter is None:
        query = layer.filter
    from arcgis.geometry import Geometry, SpatialReference

    fields = []
    records = []
    if isinstance(layer, (Table, FeatureLayer)) == False:
        raise ValueError("Invalid inputs: must be FeatureLayer or Table")
    sdf = layer.query(where=query, as_df=True)
    sdf.spatial._meta.source = layer
    if "drawingInfo" in layer.properties:
        sdf.spatial.renderer = dict(layer.properties.drawingInfo.renderer)
    else:
        sdf.spatial.renderer = dict({})
    return sdf


# ----------------------------------------------------------------------
def to_layer(df, layer, update_existing=True, add_new=False, truncate=False):
    """
    Sends the Spatial DataFrame information to a published service

    :Parameters:
     :df: Spatial DataFrame object
     :layer: Feature Layer or Table Layer object
     :update_existing: boolean -
     :add_new: boolean
     :truncate: if true, all records will be deleted and the dataframe
      records will replace the service data
    Output:
     A layer object
    """
    if not isinstance(df, (pd.DataFrame)) or not hasattr(df, "spatial"):
        raise ValueError("df must be a SpatialDatframe")
    if not isinstance(layer, (Table, FeatureLayer)):
        raise ValueError("layer must be a FeatureLayer or Table Layer")
    if truncate:
        layer.delete_features(where="1=1")
        layer.edit_features(adds=to_featureset(df).features)
    elif update_existing:
        layer.edit_features(updates=to_featureset(df).features)
    elif add_new:
        layer.edit_features(adds=to_featureset(df).features)
    return layer



def is_numpy_array(image):
    """numpy arrays behave unpredictably in `isinstance()` func calls:
    do a hacky string comparison on the type() of image arg
    """
    return "numpy" in str(type(image)) and "array" in str(type(image))


def get_hash_numpy_array(image):
    return hash(str(image.tostring()))
